"""V2.0 issue-credential anoncreds credential format handler."""

import json
import logging
from typing import List, Mapping, Optional, Tuple

from anoncreds import CredentialDefinition
from marshmallow import RAISE

from ......anoncreds.base import AnonCredsObjectNotFound, AnonCredsResolutionError
from ......anoncreds.holder import AnonCredsHolder, AnonCredsHolderError
from ......anoncreds.issuer import CATEGORY_CRED_DEF, AnonCredsIssuer
from ......anoncreds.models.credential import AnonCredsCredentialSchema
from ......anoncreds.models.credential_offer import AnonCredsCredentialOfferSchema
from ......anoncreds.models.credential_proposal import (
    AnonCredsCredentialDefinitionProposal,
)
from ......anoncreds.models.credential_request import AnonCredsCredRequestSchema
from ......anoncreds.registry import AnonCredsRegistry
from ......anoncreds.revocation import AnonCredsRevocation
from ......cache.base import BaseCache
from ......messaging.credential_definitions.util import CRED_DEF_SENT_RECORD_TYPE
from ......messaging.decorators.attach_decorator import AttachDecorator
from ......revocation_anoncreds.models.issuer_cred_rev_record import IssuerCredRevRecord
from ......storage.base import BaseStorage
from ...message_types import (
    ATTACHMENT_FORMAT,
    CRED_20_ISSUE,
    CRED_20_OFFER,
    CRED_20_PROPOSAL,
    CRED_20_REQUEST,
)
from ...messages.cred_format import V20CredFormat
from ...messages.cred_issue import V20CredIssue
from ...messages.cred_offer import V20CredOffer
from ...messages.cred_proposal import V20CredProposal
from ...messages.cred_request import V20CredRequest
from ...models.cred_ex_record import V20CredExRecord
from ...models.detail.anoncreds import V20CredExRecordAnonCreds
from ..handler import CredFormatAttachment, V20CredFormatError, V20CredFormatHandler

LOGGER = logging.getLogger(__name__)


class AnonCredsCredFormatHandler(V20CredFormatHandler):
    """AnonCreds credential format handler."""

    format = V20CredFormat.Format.ANONCREDS

    @classmethod
    def validate_fields(cls, message_type: str, attachment_data: Mapping):
        """Validate attachment data for a specific message type.

        Uses marshmallow schemas to validate if format specific attachment data
        is valid for the specified message type. Only does structural and type
        checks, does not validate if .e.g. the issuer value is valid.


        Args:
            message_type (str): The message type to validate the attachment data for.
                Should be one of the message types as defined in message_types.py
            attachment_data (Mapping): [description]
                The attachment data to validate

        Raises:
            Exception: When the data is not valid.

        """
        mapping = {
            CRED_20_PROPOSAL: AnonCredsCredentialDefinitionProposal,
            CRED_20_OFFER: AnonCredsCredentialOfferSchema,
            CRED_20_REQUEST: AnonCredsCredRequestSchema,
            CRED_20_ISSUE: AnonCredsCredentialSchema,
        }

        # Get schema class
        Schema = mapping[message_type]

        # Validate, throw if not valid
        Schema(unknown=RAISE).load(attachment_data)

    async def get_detail_record(self, cred_ex_id: str) -> V20CredExRecordAnonCreds:
        """Retrieve credential exchange detail record by cred_ex_id."""

        async with self.profile.session() as session:
            records = await AnonCredsCredFormatHandler.format.detail.query_by_cred_ex_id(
                session, cred_ex_id
            )

        if len(records) > 1:
            LOGGER.warning(
                "Cred ex id %s has %d %s detail records: should be 1",
                cred_ex_id,
                len(records),
                AnonCredsCredFormatHandler.format.api,
            )
        return records[0] if records else None

    async def _check_uniqueness(self, cred_ex_id: str):
        """Raise exception on evidence that cred ex already has cred issued to it."""
        async with self.profile.session() as session:
            exist = await AnonCredsCredFormatHandler.format.detail.query_by_cred_ex_id(
                session, cred_ex_id
            )
        if exist:
            raise V20CredFormatError(
                f"{AnonCredsCredFormatHandler.format.api} detail record already "
                f"exists for cred ex id {cred_ex_id}"
            )

    def get_format_identifier(self, message_type: str) -> str:
        """Get attachment format identifier for format and message combination.

        Args:
            message_type (str): Message type for which to return the format identifier

        Returns:
            str: Issue credential attachment format identifier

        """
        return ATTACHMENT_FORMAT[message_type][AnonCredsCredFormatHandler.format.api]

    def get_format_data(self, message_type: str, data: dict) -> CredFormatAttachment:
        """Get credential format and attachment objects for use in cred ex messages.

        Returns a tuple of both credential format and attachment decorator for use
        in credential exchange messages. It looks up the correct format identifier and
        encodes the data as a base64 attachment.

        Args:
            message_type (str): The message type for which to return the cred format.
                Should be one of the message types defined in the message types file
            data (dict): The data to include in the attach decorator

        Returns:
            CredFormatAttachment: Credential format and attachment data objects

        """
        return (
            V20CredFormat(
                attach_id=AnonCredsCredFormatHandler.format.api,
                format_=self.get_format_identifier(message_type),
            ),
            AttachDecorator.data_base64(
                data, ident=AnonCredsCredFormatHandler.format.api
            ),
        )

    async def _match_sent_cred_def_id(self, tag_query: Mapping[str, str]) -> str:
        """Return most recent matching id of cred def that agent sent to ledger."""

        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            found = await storage.find_all_records(
                type_filter=CRED_DEF_SENT_RECORD_TYPE, tag_query=tag_query
            )
        if not found:
            raise V20CredFormatError(
                f"Issuer has no operable cred def for proposal spec {tag_query}"
            )
        return max(found, key=lambda r: int(r.tags["epoch"])).tags["cred_def_id"]

    async def create_proposal(
        self, cred_ex_record: V20CredExRecord, proposal_data: Mapping[str, str]
    ) -> Tuple[V20CredFormat, AttachDecorator]:
        """Create anoncreds credential proposal."""
        if proposal_data is None:
            proposal_data = {}

        return self.get_format_data(CRED_20_PROPOSAL, proposal_data)

    async def receive_proposal(
        self, cred_ex_record: V20CredExRecord, cred_proposal_message: V20CredProposal
    ) -> None:
        """Receive anoncreds credential proposal.

        No custom handling is required for this step.
        """

    async def create_offer(
        self, cred_proposal_message: V20CredProposal
    ) -> CredFormatAttachment:
        """Create anoncreds credential offer."""

        issuer = AnonCredsIssuer(self.profile)
        cache = self.profile.inject_or(BaseCache)

        anoncreds_attachment = cred_proposal_message.attachment(
            AnonCredsCredFormatHandler.format
        )

        if not anoncreds_attachment:
            anoncreds_attachment = cred_proposal_message.attachment(
                V20CredFormat.Format.INDY.api
            )

        cred_def_id = await issuer.match_created_credential_definitions(
            **anoncreds_attachment
        )

        async def _create():
            offer_json = await issuer.create_credential_offer(cred_def_id)
            return json.loads(offer_json)

        async def _get_attr_names(schema_id) -> List[str] | None:
            """Fetch attribute names for a given schema ID from the registry."""
            if not schema_id:
                return None
            try:
                schema_result = await registry.get_schema(self.profile, schema_id)
                return schema_result.schema.attr_names
            except AnonCredsObjectNotFound:
                LOGGER.info(f"Schema not found for schema_id={schema_id}")
                return None
            except AnonCredsResolutionError as e:
                LOGGER.warning(f"Schema resolution failed for schema_id={schema_id}: {e}")
                return None

        async def _fetch_schema_attr_names(
            anoncreds_attachment, cred_def_id
        ) -> List[str] | None:
            """Determine schema attribute names from schema_id or cred_def_id."""
            schema_id = anoncreds_attachment.get("schema_id")
            attr_names = await _get_attr_names(schema_id)

            if attr_names:
                return attr_names

            if cred_def_id:
                async with self.profile.session() as session:
                    cred_def_entry = await session.handle.fetch(
                        CATEGORY_CRED_DEF, cred_def_id
                    )
                    cred_def_dict = CredentialDefinition.load(
                        cred_def_entry.value
                    ).to_dict()
                    return await _get_attr_names(cred_def_dict.get("schemaId"))

            return None

        attr_names = None
        registry = self.profile.inject(AnonCredsRegistry)

        attr_names = await _fetch_schema_attr_names(anoncreds_attachment, cred_def_id)

        if not attr_names:
            raise V20CredFormatError(
                "Could not determine schema attributes. If you did not create the "
                "schema, then you need to provide the schema_id."
            )

        schema_attrs = set(attr_names)
        preview_attrs = set(cred_proposal_message.credential_preview.attr_dict())
        if preview_attrs != schema_attrs:
            raise V20CredFormatError(
                f"Preview attributes {preview_attrs} "
                f"mismatch corresponding schema attributes {schema_attrs}"
            )

        cred_offer = None
        cache_key = f"credential_offer::{cred_def_id}"

        if cache:
            async with cache.acquire(cache_key) as entry:
                if entry.result:
                    cred_offer = entry.result
                else:
                    cred_offer = await _create()
                    await entry.set_result(cred_offer, 3600)
        if not cred_offer:
            cred_offer = await _create()

        return self.get_format_data(CRED_20_OFFER, cred_offer)

    async def receive_offer(
        self, cred_ex_record: V20CredExRecord, cred_offer_message: V20CredOffer
    ) -> None:
        """Receive anoncreds credential offer."""

    async def create_request(
        self, cred_ex_record: V20CredExRecord, request_data: Optional[Mapping] = None
    ) -> CredFormatAttachment:
        """Create anoncreds credential request."""
        if cred_ex_record.state != V20CredExRecord.STATE_OFFER_RECEIVED:
            raise V20CredFormatError(
                "AnonCreds issue credential format cannot start from credential request"
            )

        await self._check_uniqueness(cred_ex_record.cred_ex_id)
        holder_did = request_data.get("holder_did") if request_data else None

        # For backwards compatibility, remove indy backup when indy format is retired
        from ..indy.handler import IndyCredFormatHandler

        cred_offer = cred_ex_record.cred_offer.attachment(
            AnonCredsCredFormatHandler.format
        ) or cred_ex_record.cred_offer.attachment(IndyCredFormatHandler.format)

        if "nonce" not in cred_offer:
            raise V20CredFormatError("Missing nonce in credential offer")

        nonce = cred_offer["nonce"]
        cred_def_id = cred_offer["cred_def_id"]

        async def _create():
            anoncreds_registry = self.profile.inject(AnonCredsRegistry)

            try:
                cred_def_result = await anoncreds_registry.get_credential_definition(
                    self.profile, cred_def_id
                )
                holder = AnonCredsHolder(self.profile)
                request_json, metadata_json = await holder.create_credential_request(
                    cred_offer, cred_def_result.credential_definition, holder_did
                )

                return {
                    "request": json.loads(request_json),
                    "metadata": json.loads(metadata_json),
                }
            # This is for compatability with a holder that isn't anoncreds capable
            except AnonCredsResolutionError:
                return await IndyCredFormatHandler.create_cred_request_result(
                    self, cred_offer, holder_did, cred_def_id
                )

        cache_key = f"credential_request::{cred_def_id}::{holder_did}::{nonce}"
        cred_req_result = None
        cache = self.profile.inject_or(BaseCache)
        if cache:
            async with cache.acquire(cache_key) as entry:
                if entry.result:
                    cred_req_result = entry.result
                else:
                    cred_req_result = await _create()
                    await entry.set_result(cred_req_result, 3600)
        if not cred_req_result:
            cred_req_result = await _create()

        detail_record = V20CredExRecordAnonCreds(
            cred_ex_id=cred_ex_record.cred_ex_id,
            cred_request_metadata=cred_req_result["metadata"],
        )

        async with self.profile.session() as session:
            await detail_record.save(session, reason="create v2.0 credential request")

        return self.get_format_data(CRED_20_REQUEST, cred_req_result["request"])

    async def receive_request(
        self, cred_ex_record: V20CredExRecord, cred_request_message: V20CredRequest
    ) -> None:
        """Receive anoncreds credential request."""
        if not cred_ex_record.cred_offer:
            raise V20CredFormatError(
                "AnonCreds issue credential format cannot start from credential request"
            )

    async def issue_credential(
        self, cred_ex_record: V20CredExRecord, retries: int = 5
    ) -> CredFormatAttachment:
        """Issue anoncreds credential."""
        await self._check_uniqueness(cred_ex_record.cred_ex_id)

        # For backwards compatibility, remove indy backup when indy format is retired
        from ..indy.handler import IndyCredFormatHandler

        if cred_ex_record.cred_offer.attachment(IndyCredFormatHandler.format):
            indy_handler = IndyCredFormatHandler(self.profile)
            return await indy_handler.issue_credential(cred_ex_record, retries)

        cred_offer = cred_ex_record.cred_offer.attachment(
            AnonCredsCredFormatHandler.format
        )
        cred_request = cred_ex_record.cred_request.attachment(
            AnonCredsCredFormatHandler.format
        )
        cred_values = cred_ex_record.cred_offer.credential_preview.attr_dict(decode=False)

        issuer = AnonCredsIssuer(self.profile)
        cred_def_id = cred_offer["cred_def_id"]
        if await issuer.cred_def_supports_revocation(cred_def_id):
            revocation = AnonCredsRevocation(self.profile)
            cred_json, cred_rev_id, rev_reg_def_id = await revocation.create_credential(
                cred_offer, cred_request, cred_values
            )
        else:
            cred_json = await issuer.create_credential(
                cred_offer, cred_request, cred_values
            )
            cred_rev_id = None
            rev_reg_def_id = None

        result = self.get_format_data(CRED_20_ISSUE, json.loads(cred_json))

        async with self._profile.transaction() as txn:
            detail_record = V20CredExRecordAnonCreds(
                cred_ex_id=cred_ex_record.cred_ex_id,
                rev_reg_id=rev_reg_def_id,
                cred_rev_id=cred_rev_id,
            )
            await detail_record.save(txn, reason="v2.0 issue credential")

            if cred_rev_id:
                issuer_cr_rec = IssuerCredRevRecord(
                    state=IssuerCredRevRecord.STATE_ISSUED,
                    cred_ex_id=cred_ex_record.cred_ex_id,
                    cred_ex_version=IssuerCredRevRecord.VERSION_2,
                    rev_reg_id=rev_reg_def_id,
                    cred_rev_id=cred_rev_id,
                )
                await issuer_cr_rec.save(
                    txn,
                    reason=(
                        "Created issuer cred rev record for "
                        f"rev reg id {rev_reg_def_id}, index {cred_rev_id}"
                    ),
                )
            await txn.commit()

        return result

    async def receive_credential(
        self, cred_ex_record: V20CredExRecord, cred_issue_message: V20CredIssue
    ) -> None:
        """Receive anoncreds credential.

        Validation is done in the store credential step.
        """

    async def store_credential(
        self, cred_ex_record: V20CredExRecord, cred_id: Optional[str] = None
    ) -> None:
        """Store anoncreds credential."""

        # For backwards compatibility, remove indy backup when indy format is retired
        from ..indy.handler import IndyCredFormatHandler

        cred = cred_ex_record.cred_issue.attachment(
            AnonCredsCredFormatHandler.format
        ) or cred_ex_record.cred_issue.attachment(IndyCredFormatHandler.format)

        rev_reg_def = None
        anoncreds_registry = self.profile.inject(AnonCredsRegistry)
        try:
            cred_def_result = await anoncreds_registry.get_credential_definition(
                self.profile, cred["cred_def_id"]
            )
            if cred.get("rev_reg_id"):
                rev_reg_def_result = (
                    await anoncreds_registry.get_revocation_registry_definition(
                        self.profile, cred["rev_reg_id"]
                    )
                )
                rev_reg_def = rev_reg_def_result.revocation_registry

        except AnonCredsResolutionError:
            return await IndyCredFormatHandler.store_credential(
                self, cred_ex_record, cred_id
            )

        holder = AnonCredsHolder(self.profile)
        cred_offer_message = cred_ex_record.cred_offer
        mime_types = None
        if cred_offer_message and cred_offer_message.credential_preview:
            mime_types = cred_offer_message.credential_preview.mime_types() or None

        if rev_reg_def:
            revocation = AnonCredsRevocation(self.profile)
            await revocation.get_or_fetch_local_tails_path(rev_reg_def)
        try:
            detail_record = await self.get_detail_record(cred_ex_record.cred_ex_id)
            if detail_record is None:
                raise V20CredFormatError(
                    f"No credential exchange {AnonCredsCredFormatHandler.format.aries} "
                    f"detail record found for cred ex id {cred_ex_record.cred_ex_id}"
                )
            cred_id_stored = await holder.store_credential(
                cred_def_result.credential_definition.serialize(),
                cred,
                detail_record.cred_request_metadata,
                mime_types,
                credential_id=cred_id,
                rev_reg_def=rev_reg_def.serialize() if rev_reg_def else None,
            )

            detail_record.cred_id_stored = cred_id_stored
            detail_record.rev_reg_id = cred.get("rev_reg_id", None)
            detail_record.cred_rev_id = cred.get("cred_rev_id", None)

            async with self.profile.session() as session:
                # Store detail record, emit event
                await detail_record.save(
                    session, reason="store credential v2.0", event=True
                )
        except AnonCredsHolderError as e:
            LOGGER.error(f"Error storing credential: {e.error_code} - {e.message}")
            raise e



"""Endorsed transaction response handler."""

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from .....revocation.util import notify_rev_reg_entry_txn_failed
from .....storage.error import StorageError
from ..manager import TransactionManager, TransactionManagerError
from ..messages.endorsed_transaction_response import EndorsedTransactionResponse


class EndorsedTransactionResponseHandler(BaseHandler):
    """Handler class for Endorsed transaction response."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle endorsed transaction response.

        Args:
            context: Request context
            responder: Responder callback
        """

        self._logger.debug(
            "EndorsedTransactionResponseHandler called with context %s", context
        )
        assert isinstance(context.message, EndorsedTransactionResponse)

        if not context.connection_ready:
            raise HandlerException("No connection established")

        async def send_failed_transaction_event(err_msg: str):
            await notify_rev_reg_entry_txn_failed(context.profile, err_msg)

        mgr = TransactionManager(context.profile)
        try:
            transaction = await mgr.receive_endorse_response(context.message)
        except TransactionManagerError as err:
            self._logger.exception("Error receiving endorsed transaction response")
            await send_failed_transaction_event(str(err))
            raise HandlerException(str(err))

        # Automatically write transaction if flag is set
        if context.settings.get("endorser.auto_write"):
            try:
                (
                    transaction,
                    transaction_acknowledgement_message,
                ) = await mgr.complete_transaction(transaction, False)

                await responder.send_reply(
                    transaction_acknowledgement_message,
                    connection_id=transaction.connection_id,
                )
            except (StorageError, TransactionManagerError) as err:
                self._logger.exception(err)
                await send_failed_transaction_event(str(err))
                raise HandlerException(str(err))



from unittest import IsolatedAsyncioTestCase

from ...error import WalletSettingsError
from ..wallet_record import WalletRecord


class TestWalletRecord(IsolatedAsyncioTestCase):
    async def test_serde(self):
        rec = WalletRecord(
            wallet_id="my-wallet-id",
            key_management_mode=WalletRecord.MODE_MANAGED,
            settings={
                "wallet.name": "my-wallet",
                "wallet.type": "indy",
                "wallet.key": "dummy-wallet-key",
            },
            wallet_name="my-wallet",
        )
        ser = rec.serialize()
        assert ser["wallet_id"] == rec.wallet_id
        assert ser["key_management_mode"] == WalletRecord.MODE_MANAGED
        assert len(ser["settings"]) == 4

        assert rec == WalletRecord.deserialize(ser)

    async def test_rec_ops(self):
        recs = [
            WalletRecord(
                wallet_id=f"my-wallet-id-{i}",
                key_management_mode=[
                    WalletRecord.MODE_UNMANAGED,
                    WalletRecord.MODE_MANAGED,
                ][i],
                settings={
                    "wallet.name": f"my-wallet-{i}",
                    "wallet.type": "indy",
                    "wallet.key": f"dummy-wallet-key-{i}",
                },
                wallet_name=f"my-wallet-{i}",
            )
            for i in range(2)
        ]
        assert recs[0].wallet_id
        assert recs[0] != recs[1]
        assert recs[0].wallet_name
        assert recs[0].wallet_type == "indy"
        assert recs[0].wallet_key
        assert recs[0].record_value

        assert not recs[0].is_managed
        assert recs[1].is_managed

    async def test_requires_external_key_in_memory(self):
        wallet_record = WalletRecord(
            settings={"wallet.type": "in_memory"},
        )

        # should be false for in_memory wallets
        assert wallet_record.requires_external_key is False

    async def test_requires_external_key_managed(self):
        wallet_record = WalletRecord(
            wallet_id="test",
            settings={"wallet.type": "indy"},
            key_management_mode=WalletRecord.MODE_MANAGED,
        )

        # should be false for managed wallets
        assert wallet_record.requires_external_key is False

    async def test_requires_external_key_unmanaged(self):
        wallet_record = WalletRecord(
            wallet_id="test",
            settings={"wallet.type": "indy"},
            key_management_mode=WalletRecord.MODE_UNMANAGED,
        )

        # should return true if wallet is unmanaged and wallet_type != unmanaged
        assert wallet_record.requires_external_key is True

    async def test_update_settings(self):
        wallet_record = WalletRecord(
            settings={"wallet.type": "in_memory"},
        )
        settings = {
            "wallet.type": "indy",
        }
        wallet_record.update_settings(settings)

        assert wallet_record.settings.get("wallet.type") == "indy"

    async def test_update_settings_x(self):
        wallet_record = WalletRecord()
        settings = {
            "wallet.id": "my-wallet-id",
        }
        with self.assertRaises(WalletSettingsError):
            wallet_record.update_settings(settings)



"""Routing manager classes for tracking and inspecting routing records."""

import asyncio
import logging
from typing import Optional, Sequence

from ....core.error import BaseError
from ....core.profile import Profile
from ....storage.error import StorageDuplicateError, StorageNotFoundError
from .models.route_record import RouteRecord

LOGGER = logging.getLogger(__name__)

RECIP_ROUTE_PAUSE = 0.1
RECIP_ROUTE_RETRY = 10


class RoutingManagerError(BaseError):
    """Generic routing error."""


class RouteNotFoundError(RoutingManagerError):
    """Requested route was not found."""


class RoutingManager:
    """Class for handling routing records."""

    RECORD_TYPE = "forward_route"

    def __init__(self, profile: Profile):
        """Initialize a RoutingManager.

        Args:
            profile: The profile instance for this manager
        """
        self._profile = profile
        if not profile:
            raise RoutingManagerError("Missing profile")

    async def get_recipient(self, recip_verkey: str) -> RouteRecord:
        """Resolve the recipient for a verkey.

        Args:
            recip_verkey: The verkey ("to") of the incoming Forward message

        Returns:
            The `RouteRecord` associated with this verkey

        """
        if not recip_verkey:
            raise RoutingManagerError("Must pass non-empty recip_verkey")

        i = 0
        record = None
        while not record:
            try:
                LOGGER.debug("Fetching routing record for verkey: %s", recip_verkey)
                async with self._profile.session() as session:
                    record = await RouteRecord.retrieve_by_recipient_key(
                        session, recip_verkey
                    )
                LOGGER.debug("Found routing record for verkey: %s", recip_verkey)
                return record
            except StorageDuplicateError:
                LOGGER.info(
                    "Duplicate routing records found for verkey: %s", recip_verkey
                )
                raise RouteNotFoundError(
                    f"More than one route record found with recipient key: {recip_verkey}"
                )
            except StorageNotFoundError:
                LOGGER.debug("No routing record found for verkey: %s", recip_verkey)
                i += 1
                if i > RECIP_ROUTE_RETRY:
                    raise RouteNotFoundError(
                        f"No route found with recipient key: {recip_verkey}"
                    )
                await asyncio.sleep(RECIP_ROUTE_PAUSE)

    async def get_routes(
        self,
        client_connection_id: Optional[str] = None,
        tag_filter: Optional[dict] = None,
    ) -> Sequence[RouteRecord]:
        """Fetch all routes associated with the current connection.

        Args:
            client_connection_id: The ID of the connection record
            tag_filter: An optional dictionary of tag filters

        Returns:
            A sequence of route records found by the query

        """
        # Routing protocol acts only as Server, filter out all client records
        filters = {"role": RouteRecord.ROLE_SERVER}
        if client_connection_id:
            filters["connection_id"] = client_connection_id
        if tag_filter:
            for key in ("recipient_key",):
                if key not in tag_filter:
                    continue
                val = tag_filter[key]
                if isinstance(val, str):
                    filters[key] = val
                elif isinstance(val, list):
                    filters[key] = {"$in": val}
                else:
                    raise RoutingManagerError(
                        "Unsupported tag filter: '{}' = {}".format(key, val)
                    )

        async with self._profile.session() as session:
            results = await RouteRecord.query(session, tag_filter=filters)

        return results

    async def delete_route_record(self, route: RouteRecord):
        """Remove an existing route record."""
        async with self._profile.session() as session:
            await route.delete_record(session)

    async def create_route_record(
        self,
        client_connection_id: Optional[str] = None,
        recipient_key: Optional[str] = None,
        internal_wallet_id: Optional[str] = None,
    ) -> RouteRecord:
        """Create and store a new RouteRecord.

        Args:
            client_connection_id: The ID of the connection record
            recipient_key: The recipient verkey of the route
            internal_wallet_id: The ID of the wallet record. Used for internal routing

        Returns:
            The new routing record

        """
        if not (client_connection_id or internal_wallet_id):
            raise RoutingManagerError(
                "Either client_connection_id or internal_wallet_id is required"
            )
        if not recipient_key:
            raise RoutingManagerError("Missing recipient_key")
        LOGGER.debug("Creating routing record for verkey: %s", recipient_key)
        route = RouteRecord(
            connection_id=client_connection_id,
            wallet_id=internal_wallet_id,
            recipient_key=recipient_key,
        )
        async with self._profile.session() as session:
            await route.save(session, reason="Created new route")
        LOGGER.info("Created routing record for verkey: %s", recipient_key)
        return route



DID_EXAMPLE_48939859 = {
    "@context": "https://www.w3.org/ns/did/v1",
    "id": "did:example:489398593",
    "assertionMethod": [
        {
            "id": "did:example:489398593#test",
            "type": "Bls12381G2Key2020",
            "controller": "did:example:489398593",
            "publicKeyBase58": "oqpWYKaZD9M1Kbe94BVXpr8WTdFBNZyKv48cziTiQUeuhm7sBhCABMyYG4kcMrseC68YTFFgyhiNeBKjzdKk9MiRWuLv5H4FFujQsQK2KTAtzU8qTBiZqBHMmnLF4PL7Ytu",
        }
    ],
}



"""Test DIDWebRegistry."""

import re

import pytest

from ..registry import DIDWebRegistry

DID_WEB = re.compile(
    r"^did:web:[a-z0-9]+(?:\.[a-z0-9]+)*(?::\d+)?(?:\/[^#\s]*)?(?:#.*)?\s*$"
)


TEST_WED_DID_0 = "did:web:example.com/anoncreds/v0/SCHEMA/asdf"
TEST_WED_DID_1 = "did:web:example.com"
TEST_WED_DID_2 = "did:web:sub.example.com"
TEST_WED_DID_3 = "did:web:example.com:8080"
TEST_WED_DID_4 = "did:web:sub.example.com/path/to/resource"
TEST_WED_DID_5 = "did:web:example.com/path/to/resource#fragment"


@pytest.fixture
def registry():
    """Registry fixture"""
    yield DIDWebRegistry()


class TestLegacyIndyRegistry:
    @pytest.mark.asyncio
    async def test_supported_did_regex(self, registry: DIDWebRegistry):
        """Test the supported_did_regex."""

        assert registry.supported_identifiers_regex == DID_WEB
        assert bool(registry.supported_identifiers_regex.match(TEST_WED_DID_0))
        assert bool(registry.supported_identifiers_regex.match(TEST_WED_DID_1))
        assert bool(registry.supported_identifiers_regex.match(TEST_WED_DID_2))
        assert bool(registry.supported_identifiers_regex.match(TEST_WED_DID_3))
        assert bool(registry.supported_identifiers_regex.match(TEST_WED_DID_4))
        assert bool(registry.supported_identifiers_regex.match(TEST_WED_DID_5))



"""Utils for messages."""

import logging
import re
from datetime import datetime, timedelta, timezone
from hashlib import sha256
from math import floor
from typing import Any, Dict, List, Union

LOGGER = logging.getLogger(__name__)
I32_BOUND = 2**31


def datetime_to_str(dt: Union[str, datetime, None]) -> Union[str, None]:
    """Convert a datetime object to an indy-standard datetime string.

    Args:
        dt: May be a string or datetime to allow automatic conversion
    """
    if isinstance(dt, datetime):
        dt = dt.replace(tzinfo=timezone.utc).isoformat().replace("+00:00", "Z")
    return dt


def str_to_datetime(dt: Union[str, datetime]) -> datetime:
    """Convert an indy-standard datetime string to a datetime.

    Using a fairly lax regex pattern to match slightly different formats.
    In Python 3.7 datetime.fromisoformat might be used.

    Args:
        dt: May be a string or datetime to allow automatic conversion

    """
    if isinstance(dt, str):
        match = re.match(
            r"^(\d{4})-(\d\d)-(\d\d)[T ](\d\d):(\d\d)"
            r"(?:\:(\d\d(?:\.\d+)?))?([+-]\d\d:?\d\d|Z|)$",
            dt,
        )
        if not match:
            raise ValueError("String does not match expected time format")
        year, month, day = match[1], match[2], match[3]
        hour, minute, second = match[4], match[5], match[6]
        tz = match[7]
        if second:
            flt_second = float(second)
            second = floor(flt_second)
            microsecond = round((flt_second - second) * 1_000_000)
        else:
            second = 0
            microsecond = 0
        result = datetime(
            int(year),
            int(month),
            int(day),
            int(hour),
            int(minute),
            int(second),
            microsecond,
            timezone.utc,
        )
        if tz != "Z" and tz != "":
            tz_sgn = int(tz[0] + "1")
            tz_hours = int(tz[1:3])
            tz_mins = int(tz[-2:])
            if tz_hours or tz_mins:
                result = result - timedelta(minutes=tz_sgn * (tz_hours * 60 + tz_mins))
        return result
    return dt


def str_to_epoch(dt: Union[str, datetime]) -> int:
    """Convert an indy-standard datetime string to epoch seconds.

    Args:
        dt: May be a string or datetime to allow automatic conversion

    """
    return int(str_to_datetime(dt).timestamp())


def epoch_to_str(epoch: int) -> str:
    """Convert epoch seconds to indy-standard datetime string.

    Args:
        epoch: epoch seconds

    """
    return datetime_to_str(datetime.fromtimestamp(epoch, tz=timezone.utc))


def datetime_now() -> datetime:
    """Timestamp in UTC."""
    return datetime.now(tz=timezone.utc)


def time_now() -> str:
    """Timestamp in ISO format."""
    return datetime_to_str(datetime_now())


def encode(orig: Any) -> str:
    """Encode a credential value as an int.

    Encode credential attribute value, purely stringifying any int32
    and leaving numeric int32 strings alone, but mapping any other
    input to a stringified 256-bit (but not 32-bit) integer.
    Predicates in indy-sdk operate
    on int32 values properly only when their encoded values match their raw values.

    Args:
        orig: original value to encode

    Returns:
        encoded value

    """

    if isinstance(orig, int) and -I32_BOUND <= orig < I32_BOUND:
        return str(int(orig))  # python bools are ints

    try:
        i32orig = int(str(orig))  # don't encode floats as ints
        if -I32_BOUND <= i32orig < I32_BOUND:
            return str(i32orig)
    except (ValueError, TypeError):
        pass

    rv = int.from_bytes(sha256(str(orig).encode()).digest(), "big")

    return str(rv)


def canon(raw_attr_name: str) -> str:
    """Canonicalize input attribute name for indy proofs and credential offers.

    Args:
        raw_attr_name: raw attribute name

    Returns:
        canonicalized attribute name

    """
    if raw_attr_name:  # do not dereference None, and "" is already canonical
        return raw_attr_name.replace(" ", "").lower()
    return raw_attr_name


def get_proto_default_version(
    versions: List[Dict[str, Any]], major_version: int = 1
) -> str:
    """Return default protocol version from version definition list."""

    for version in versions:
        if major_version == version["major_version"]:
            default_major_version = version["major_version"]
            default_minor_version = version["current_minor_version"]
            return f"{default_major_version}.{default_minor_version}"

    return "1.0"



"""Classes to manage presentations."""

import json
import logging
from typing import Optional

from ....connections.models.conn_record import ConnRecord
from ....core.error import BaseError
from ....core.profile import Profile
from ....indy.verifier import IndyVerifier
from ....messaging.decorators.attach_decorator import AttachDecorator
from ....messaging.responder import BaseResponder
from ....storage.error import StorageNotFoundError
from ...out_of_band.v1_0.models.oob_record import OobRecord
from ..indy.pres_exch_handler import IndyPresExchHandler
from .message_types import ATTACH_DECO_IDS, PRESENTATION, PRESENTATION_REQUEST
from .messages.presentation import Presentation
from .messages.presentation_ack import PresentationAck
from .messages.presentation_problem_report import (
    PresentationProblemReport,
    ProblemReportReason,
)
from .messages.presentation_proposal import PresentationProposal
from .messages.presentation_request import PresentationRequest
from .models.presentation_exchange import V10PresentationExchange

LOGGER = logging.getLogger(__name__)


class PresentationManagerError(BaseError):
    """Presentation error."""


class PresentationManager:
    """Class for managing presentations."""

    def __init__(self, profile: Profile):
        """Initialize a PresentationManager.

        Args:
            profile: The profile instance for this presentation manager
        """

        self._profile = profile

    async def create_exchange_for_proposal(
        self,
        connection_id: str,
        presentation_proposal_message: PresentationProposal,
        auto_present: Optional[bool] = None,
        auto_remove: Optional[bool] = None,
    ):
        """Create a presentation exchange record for input presentation proposal.

        Args:
            connection_id: connection identifier
            presentation_proposal_message: presentation proposal to serialize
                to exchange record
            auto_present: whether to present proof upon receiving proof request
                (default to configuration setting)
            auto_remove: whether to remove this presentation exchange upon completion

        Returns:
            Presentation exchange record, created

        """
        if auto_remove is None:
            auto_remove = not self._profile.settings.get("preserve_exchange_records")
        presentation_exchange_record = V10PresentationExchange(
            connection_id=connection_id,
            thread_id=presentation_proposal_message._thread_id,
            initiator=V10PresentationExchange.INITIATOR_SELF,
            role=V10PresentationExchange.ROLE_PROVER,
            state=V10PresentationExchange.STATE_PROPOSAL_SENT,
            presentation_proposal_dict=presentation_proposal_message,
            auto_present=auto_present,
            trace=(presentation_proposal_message._trace is not None),
            auto_remove=auto_remove,
        )
        async with self._profile.session() as session:
            await presentation_exchange_record.save(
                session, reason="create presentation proposal"
            )

        return presentation_exchange_record

    async def receive_proposal(
        self, message: PresentationProposal, connection_record: ConnRecord
    ):
        """Receive a presentation proposal from message in context on manager creation.

        Returns:
            Presentation exchange record, created

        """
        presentation_exchange_record = V10PresentationExchange(
            connection_id=connection_record.connection_id,
            thread_id=message._thread_id,
            initiator=V10PresentationExchange.INITIATOR_EXTERNAL,
            role=V10PresentationExchange.ROLE_VERIFIER,
            state=V10PresentationExchange.STATE_PROPOSAL_RECEIVED,
            presentation_proposal_dict=message,
            trace=(message._trace is not None),
            auto_remove=not self._profile.settings.get("preserve_exchange_records"),
        )
        async with self._profile.session() as session:
            await presentation_exchange_record.save(
                session, reason="receive presentation request"
            )

        return presentation_exchange_record

    async def create_bound_request(
        self,
        presentation_exchange_record: V10PresentationExchange,
        name: Optional[str] = None,
        version: Optional[str] = None,
        nonce: Optional[str] = None,
        comment: Optional[str] = None,
    ):
        """Create a presentation request bound to a proposal.

        Args:
            presentation_exchange_record: Presentation exchange record for which
                to create presentation request
            name: name to use in presentation request (None for default)
            version: version to use in presentation request (None for default)
            nonce: nonce to use in presentation request (None to generate)
            comment: Optional human-readable comment pertaining to request creation

        Returns:
            A tuple (updated presentation exchange record, presentation request message)

        """
        indy_proof_request = await (
            presentation_exchange_record.presentation_proposal_dict
        ).presentation_proposal.indy_proof_request(
            name=name,
            version=version,
            nonce=nonce,
            profile=self._profile,
        )
        presentation_request_message = PresentationRequest(
            comment=comment,
            request_presentations_attach=[
                AttachDecorator.data_base64(
                    mapping=indy_proof_request,
                    ident=ATTACH_DECO_IDS[PRESENTATION_REQUEST],
                )
            ],
        )
        presentation_request_message._thread = {
            "thid": presentation_exchange_record.thread_id
        }
        presentation_request_message.assign_trace_decorator(
            self._profile.settings, presentation_exchange_record.trace
        )

        presentation_exchange_record.thread_id = presentation_request_message._thread_id
        presentation_exchange_record.state = V10PresentationExchange.STATE_REQUEST_SENT
        presentation_exchange_record.presentation_request = indy_proof_request
        async with self._profile.session() as session:
            await presentation_exchange_record.save(
                session, reason="create (bound) presentation request"
            )

        return presentation_exchange_record, presentation_request_message

    async def create_exchange_for_request(
        self,
        connection_id: str,
        presentation_request_message: PresentationRequest,
        auto_verify: Optional[bool] = None,
        auto_remove: Optional[bool] = None,
    ):
        """Create a presentation exchange record for input presentation request.

        Args:
            connection_id: connection identifier
            presentation_request_message: presentation request to use in creating
                exchange record, extracting indy proof request and thread id
            auto_verify: whether to auto-verify presentation exchange
            auto_remove: whether to remove this presentation exchange upon completion
        Returns:
            Presentation exchange record, updated

        """
        if auto_remove is None:
            auto_remove = not self._profile.settings.get("preserve_exchange_records")
        presentation_exchange_record = V10PresentationExchange(
            connection_id=connection_id,
            thread_id=presentation_request_message._thread_id,
            initiator=V10PresentationExchange.INITIATOR_SELF,
            role=V10PresentationExchange.ROLE_VERIFIER,
            state=V10PresentationExchange.STATE_REQUEST_SENT,
            presentation_request=presentation_request_message.indy_proof_request(),
            presentation_request_dict=presentation_request_message,
            auto_verify=auto_verify,
            trace=(presentation_request_message._trace is not None),
            auto_remove=auto_remove,
        )
        async with self._profile.session() as session:
            await presentation_exchange_record.save(
                session, reason="create (free) presentation request"
            )

        return presentation_exchange_record

    async def receive_request(
        self, presentation_exchange_record: V10PresentationExchange
    ):
        """Receive a presentation request.

        Args:
            presentation_exchange_record: presentation exchange record with
                request to receive

        Returns:
            The presentation_exchange_record, updated

        """
        presentation_exchange_record.state = (
            V10PresentationExchange.STATE_REQUEST_RECEIVED
        )
        async with self._profile.session() as session:
            await presentation_exchange_record.save(
                session, reason="receive presentation request"
            )

        return presentation_exchange_record

    async def create_presentation(
        self,
        presentation_exchange_record: V10PresentationExchange,
        requested_credentials: dict,
        comment: Optional[str] = None,
    ):
        """Create a presentation.

        Args:
            presentation_exchange_record: Record to update
            requested_credentials: Indy formatted requested_credentials
            comment: optional human-readable comment


        Example `requested_credentials` format, mapping proof request referents (uuid)
        to wallet referents (cred id):

        ::

            {
                "self_attested_attributes": {
                    "j233ffbc-bd35-49b1-934f-51e083106f6d": "value"
                },
                "requested_attributes": {
                    "6253ffbb-bd35-49b3-934f-46e083106f6c": {
                        "cred_id": "5bfa40b7-062b-4ae0-a251-a86c87922c0e",
                        "revealed": true
                    }
                },
                "requested_predicates": {
                    "bfc8a97d-60d3-4f21-b998-85eeabe5c8c0": {
                        "cred_id": "5bfa40b7-062b-4ae0-a251-a86c87922c0e"
                    }
                }
            }

        Returns:
            A tuple (updated presentation exchange record, presentation message)

        """
        indy_handler = IndyPresExchHandler(self._profile)
        indy_proof = await indy_handler.return_presentation(
            pres_ex_record=presentation_exchange_record,
            requested_credentials=requested_credentials,
        )

        presentation_message = Presentation(
            comment=comment,
            presentations_attach=[
                AttachDecorator.data_base64(
                    mapping=indy_proof, ident=ATTACH_DECO_IDS[PRESENTATION]
                )
            ],
        )

        # Assign thid (and optionally pthid) to message
        presentation_message.assign_thread_from(
            presentation_exchange_record.presentation_request_dict
        )
        presentation_message.assign_trace_decorator(
            self._profile.settings, presentation_exchange_record.trace
        )

        # save presentation exchange state
        presentation_exchange_record.state = (
            V10PresentationExchange.STATE_PRESENTATION_SENT
        )
        presentation_exchange_record.presentation = indy_proof
        async with self._profile.session() as session:
            await presentation_exchange_record.save(session, reason="create presentation")

        return presentation_exchange_record, presentation_message

    async def receive_presentation(
        self,
        message: Presentation,
        connection_record: Optional[ConnRecord],
        oob_record: Optional[OobRecord],
    ):
        """Receive a presentation, from message in context on manager creation.

        Returns:
            presentation exchange record, retrieved and updated

        """
        presentation = message.indy_proof()

        thread_id = message._thread_id

        # Normally we only set the connection_id to None if an oob record is present
        # But present proof supports the old-style AIP-1 connectionless exchange that
        # bypasses the oob record. So we can't verify if an oob record is associated with
        # the exchange because it is possible that there is None
        #
        # A connectionless proof doesn't have a connection_id, so default to None
        # even if there is no oob record.
        if connection_record and connection_record.connection_id and not oob_record:
            connection_id = connection_record.connection_id
        else:
            connection_id = None

        async with self._profile.session() as session:
            # Find by thread_id and role. Verify connection id later
            presentation_exchange_record = (
                await V10PresentationExchange.retrieve_by_tag_filter(
                    session,
                    {"thread_id": thread_id},
                    {
                        "role": V10PresentationExchange.ROLE_VERIFIER,
                        "connection_id": connection_id,
                    },
                )
            )

        # Save connection id (if it wasn't already present)
        if connection_record:
            presentation_exchange_record.connection_id = connection_record.connection_id

        # Check for bait-and-switch in presented attribute values vs. proposal
        if presentation_exchange_record.presentation_proposal_dict:
            exchange_pres_proposal = (
                presentation_exchange_record.presentation_proposal_dict
            )
            presentation_preview = exchange_pres_proposal.presentation_proposal

            proof_req = presentation_exchange_record._presentation_request.ser
            for reft, attr_spec in presentation["requested_proof"][
                "revealed_attrs"
            ].items():
                name = proof_req["requested_attributes"][reft]["name"]
                value = attr_spec["raw"]
                if not presentation_preview.has_attr_spec(
                    cred_def_id=presentation["identifiers"][attr_spec["sub_proof_index"]][
                        "cred_def_id"
                    ],
                    name=name,
                    value=value,
                ):
                    presentation_exchange_record.state = (
                        V10PresentationExchange.STATE_ABANDONED
                    )
                    async with self._profile.session() as session:
                        await presentation_exchange_record.save(
                            session,
                            reason=(
                                f"Presentation {name}={value} mismatches proposal value"
                            ),
                        )
                    raise PresentationManagerError(
                        f"Presentation {name}={value} mismatches proposal value"
                    )

        presentation_exchange_record.presentation = presentation
        presentation_exchange_record.state = (
            V10PresentationExchange.STATE_PRESENTATION_RECEIVED
        )

        async with self._profile.session() as session:
            await presentation_exchange_record.save(
                session, reason="receive presentation"
            )

        return presentation_exchange_record

    async def verify_presentation(
        self,
        presentation_exchange_record: V10PresentationExchange,
        responder: Optional[BaseResponder] = None,
    ):
        """Verify a presentation.

        Args:
            presentation_exchange_record: presentation exchange record
                with presentation request and presentation to verify
            responder: responder to use

        Returns:
            presentation record, updated

        """
        indy_proof_request = presentation_exchange_record._presentation_request.ser
        indy_proof = presentation_exchange_record._presentation.ser
        indy_handler = IndyPresExchHandler(self._profile)
        (
            schemas,
            cred_defs,
            rev_reg_defs,
            rev_reg_entries,
        ) = await indy_handler.process_pres_identifiers(indy_proof["identifiers"])

        verifier = self._profile.inject(IndyVerifier)
        (verified_bool, verified_msgs) = await verifier.verify_presentation(
            dict(
                indy_proof_request
            ),  # copy to avoid changing the proof req in the stored pres exch
            indy_proof,
            schemas,
            cred_defs,
            rev_reg_defs,
            rev_reg_entries,
        )
        presentation_exchange_record.verified = json.dumps(verified_bool)
        presentation_exchange_record.verified_msgs = list(set(verified_msgs))
        presentation_exchange_record.state = V10PresentationExchange.STATE_VERIFIED

        async with self._profile.session() as session:
            await presentation_exchange_record.save(session, reason="verify presentation")

        await self.send_presentation_ack(presentation_exchange_record, responder)
        return presentation_exchange_record

    async def send_presentation_ack(
        self,
        presentation_exchange_record: V10PresentationExchange,
        responder: Optional[BaseResponder] = None,
    ):
        """Send acknowledgement of presentation receipt.

        Args:
            presentation_exchange_record: presentation exchange record with thread id
            responder: Responder to use

        """
        responder = responder or self._profile.inject_or(BaseResponder)

        if not presentation_exchange_record.connection_id:
            # Find associated oob record. If this presentation exchange is created
            # without oob (aip1 style connectionless) we can't send a presentation ack
            # because we don't have their service
            try:
                async with self._profile.session() as session:
                    await OobRecord.retrieve_by_tag_filter(
                        session,
                        {"attach_thread_id": presentation_exchange_record.thread_id},
                    )
            except StorageNotFoundError:
                # This can happen in AIP1 style connectionless exchange. ACA-PY only
                # supported this for receiving a presentation
                LOGGER.error(
                    "Unable to send connectionless presentation ack without associated "
                    "oob record. This can happen if proof request was sent without "
                    "wrapping it in an out of band invitation (AIP1-style)."
                )
                return

        if responder:
            presentation_ack_message = PresentationAck(
                verification_result=presentation_exchange_record.verified
            )
            presentation_ack_message._thread = {
                "thid": presentation_exchange_record.thread_id
            }
            presentation_ack_message.assign_trace_decorator(
                self._profile.settings, presentation_exchange_record.trace
            )

            await responder.send_reply(
                presentation_ack_message,
                # connection_id can be none in case of connectionless
                connection_id=presentation_exchange_record.connection_id,
            )

            # all done: delete
            if presentation_exchange_record.auto_remove:
                async with self._profile.session() as session:
                    await presentation_exchange_record.delete_record(session)
        else:
            LOGGER.warning(
                "Configuration has no BaseResponder: cannot ack presentation on %s",
                presentation_exchange_record.thread_id,
            )

    async def receive_presentation_ack(
        self, message: PresentationAck, connection_record: Optional[ConnRecord]
    ):
        """Receive a presentation ack, from message in context on manager creation.

        Returns:
            presentation exchange record, retrieved and updated

        """
        connection_id = connection_record.connection_id if connection_record else None

        async with self._profile.session() as session:
            (
                presentation_exchange_record
            ) = await V10PresentationExchange.retrieve_by_tag_filter(
                session,
                {"thread_id": message._thread_id},
                {
                    # connection_id can be null in connectionless
                    "connection_id": connection_id,
                    "role": V10PresentationExchange.ROLE_PROVER,
                },
            )
            presentation_exchange_record.verified = message._verification_result
            presentation_exchange_record.state = (
                V10PresentationExchange.STATE_PRESENTATION_ACKED
            )

            await presentation_exchange_record.save(
                session, reason="receive presentation ack"
            )

            # all done: delete
            if presentation_exchange_record.auto_remove:
                async with self._profile.session() as session:
                    await presentation_exchange_record.delete_record(session)

        return presentation_exchange_record

    async def receive_problem_report(
        self, message: PresentationProblemReport, connection_id: str
    ):
        """Receive problem report.

        Returns:
            presentation exchange record, retrieved and updated

        """
        # FIXME use transaction, fetch for_update
        async with self._profile.session() as session:
            pres_ex_record = await V10PresentationExchange.retrieve_by_tag_filter(
                session,
                {"thread_id": message._thread_id},
                {"connection_id": connection_id},
            )

            pres_ex_record.state = V10PresentationExchange.STATE_ABANDONED
            code = message.description.get("code", ProblemReportReason.ABANDONED.value)
            pres_ex_record.error_msg = f"{code}: {message.description.get('en', code)}"
            await pres_ex_record.save(session, reason="received problem report")

        return pres_ex_record



"""Message type identifiers for Routing."""

from ...didcomm_prefix import DIDCommPrefix

# Message types
FORWARD = "routing/1.0/forward"

PROTOCOL_PACKAGE = "acapy_agent.protocols.routing.v1_0"

MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {
        FORWARD: f"{PROTOCOL_PACKAGE}.messages.forward.Forward",
    }
)



"""Recover a revocation registry."""

import hashlib
import logging
import time

import aiohttp
import base58
import indy_vdr
from anoncreds import RevocationRegistry, RevocationRegistryDefinition

from ...models.revocation import RevList

LOGGER = logging.getLogger(__name__)


"""
This module calculates a new ledger accumulator, based on the revocation status
on the ledger vs revocations recorded in the wallet.
The calculated transaction can be written to the ledger to get the ledger back
in sync with the wallet.
This function can be used if there were previous revocation errors (i.e. the
credential revocation was successfully written to the wallet but the ledger write
failed.)
"""


class RevocRecoveryException(Exception):
    """Raise exception generating the recovery transaction."""


async def _check_tails_hash_for_inconsistency(tails_location: str, tails_hash: str):
    async with aiohttp.ClientSession() as session:
        LOGGER.debug("Tails URL: %s", tails_location)
        tails_data_http_response = await session.get(tails_location)
        tails_data = await tails_data_http_response.read()
        remote_tails_hash = base58.b58encode(hashlib.sha256(tails_data).digest()).decode(
            "utf-8"
        )
        if remote_tails_hash != tails_hash:
            raise RevocRecoveryException(
                f"Tails hash mismatch {remote_tails_hash} {tails_hash}"
            )
        else:
            LOGGER.debug(f"Checked tails hash: {tails_hash}")


async def fetch_txns(
    genesis_txns: str, registry_id: str, issuer_id: str
) -> tuple[
    dict,
    set[int],
]:
    """Fetch tails file and revocation registry information."""

    LOGGER.debug(f"Fetch revocation registry def {registry_id} from ledger")
    revoc_reg_delta_request = indy_vdr.ledger.build_get_revoc_reg_def_request(
        None, registry_id
    )

    pool = await indy_vdr.open_pool(transactions=genesis_txns)
    result = await pool.submit_request(revoc_reg_delta_request)
    if not result["data"]:
        raise RevocRecoveryException(f"Registry definition not found for {registry_id}")

    # Load the anoncreds revocation registry definition
    rev_reg_def_raw = result["data"]
    rev_reg_def_raw["ver"] = "1.0"
    rev_reg_def_raw["issuerId"] = issuer_id
    revoc_reg_def = RevocationRegistryDefinition.load(rev_reg_def_raw)

    await _check_tails_hash_for_inconsistency(
        revoc_reg_def.tails_location, revoc_reg_def.tails_hash
    )

    LOGGER.debug(f"Fetch revocation registry delta {registry_id} from ledger")
    to_timestamp = int(time.time())
    revoc_reg_delta_request = indy_vdr.ledger.build_get_revoc_reg_delta_request(
        None, registry_id, None, to_timestamp
    )
    result = await pool.submit_request(revoc_reg_delta_request)
    if not result["data"]:
        raise RevocRecoveryException("Error fetching delta from ledger")

    registry_from_ledger = result["data"]["value"]["accum_to"]
    registry_from_ledger["ver"] = "1.0"
    revoked = set(result["data"]["value"]["revoked"])
    LOGGER.debug("Ledger revoked indexes: %s", revoked)

    return registry_from_ledger, revoked


async def generate_ledger_rrrecovery_txn(genesis_txns: str, rev_list: RevList) -> dict:
    """Generate a new ledger accum entry, using the wallet value if revocations ahead of ledger."""  # noqa: E501

    registry_from_ledger, prev_revoked = await fetch_txns(
        genesis_txns, rev_list.rev_reg_def_id, rev_list.issuer_id
    )

    set_revoked = {
        index for index, value in enumerate(rev_list.revocation_list) if value == 1
    }
    mismatch = prev_revoked - set_revoked
    if mismatch:
        LOGGER.warning(
            "Credential index(es) revoked on the ledger, but not in wallet: %s",
            mismatch,
        )

    updates = set_revoked - prev_revoked
    if not updates:
        LOGGER.debug("No updates to perform")
        return {}
    else:
        LOGGER.debug("New revoked indexes: %s", updates)

        # Prepare the transaction to write to the ledger
        registry = RevocationRegistry.load(registry_from_ledger)
        registry = registry.to_dict()
        registry["ver"] = "1.0"
        registry["value"]["prevAccum"] = registry_from_ledger["value"]["accum"]
        registry["value"]["accum"] = rev_list.current_accumulator
        registry["value"]["issued"] = []
        registry["value"]["revoked"] = list(updates)
    return registry



import pytest

from ..multiformats import multibase, multicodec


def test_encode_decode():
    value = b"Hello World!"
    encoded = multibase.encode(value, "base58btc")
    assert encoded == "z2NEpo7TZRRrLZSi2U"
    decoded = multibase.decode(encoded)
    assert decoded == value


def test_encode_decode_by_encoding():
    value = b"Hello World!"
    encoded = multibase.encode(value, multibase.Encoding.base58btc)
    assert encoded == "z2NEpo7TZRRrLZSi2U"
    decoded = multibase.decode(encoded)
    assert decoded == value


def test_x_unknown_encoding():
    with pytest.raises(ValueError):
        multibase.encode(b"Hello World!", "fancy-encoding")


def test_x_unknown_character():
    with pytest.raises(ValueError):
        multibase.decode("fHello World!")


def test_x_invalid_encoding():
    with pytest.raises(TypeError):
        multibase.encode(b"Hello World!", 123)


def test_wrap_unwrap():
    value = b"Hello World!"
    wrapped = multicodec.wrap("ed25519-pub", value)
    codec, unwrapped = multicodec.unwrap(wrapped)
    assert codec == multicodec.multicodec("ed25519-pub")
    assert unwrapped == value


def test_wrap_unwrap_custom():
    value = b"Hello World!"
    my_codec = multicodec.Multicodec("my-codec", b"\x00\x01")
    wrapped = multicodec.wrap(my_codec, value)
    codec, unwrapped = multicodec.unwrap(wrapped, my_codec)
    assert codec == my_codec
    assert unwrapped == value


def test_wrap_unwrap_by_codec():
    value = b"Hello World!"
    wrapped = multicodec.wrap(multicodec.multicodec("ed25519-pub"), value)
    codec, unwrapped = multicodec.unwrap(wrapped, multicodec.multicodec("ed25519-pub"))
    assert codec == multicodec.multicodec("ed25519-pub")
    assert unwrapped == value


def test_x_unknown_multicodec():
    with pytest.raises(ValueError):
        multicodec.wrap("fancy-multicodec", b"Hello World!")


def test_x_invalid_multicodec():
    with pytest.raises(TypeError):
        multicodec.wrap(123, b"Hello World!")


def test_x_invalid_multicodec_unwrap():
    with pytest.raises(ValueError):
        multicodec.unwrap(b"Hello World!")



"""Handler for keylist-update messages."""

from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageNotFoundError
from ..manager import MediationManager, MediationNotGrantedError
from ..messages.keylist_update import KeylistUpdate
from ..messages.problem_report import CMProblemReport, ProblemReportReason
from ..models.mediation_record import MediationRecord


class KeylistUpdateHandler(BaseHandler):
    """Handler for keylist-update messages."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle keylist-update messages."""
        self._logger.debug("%s called with context %s", self.__class__.__name__, context)
        assert isinstance(context.message, KeylistUpdate)

        if not context.connection_ready:
            raise HandlerException("Cannot update routes: no active connection")

        profile = context.profile
        mgr = MediationManager(profile)
        try:
            async with profile.session() as session:
                record = await MediationRecord.retrieve_by_connection_id(
                    session, context.connection_record.connection_id
                )
            response = await mgr.update_keylist(record, updates=context.message.updates)
            response.assign_thread_from(context.message)
            await responder.send_reply(response)
        except (StorageNotFoundError, MediationNotGrantedError):
            reply = CMProblemReport(
                description={
                    "en": "Mediation has not been granted for this connection",
                    "code": ProblemReportReason.MEDIATION_NOT_GRANTED.value,
                }
            )
            reply.assign_thread_from(context.message)
            await responder.send_reply(reply)



"""Handler for mediate-grant message."""

from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....multitenant.base import BaseMultitenantManager
from .....storage.error import StorageNotFoundError
from ..manager import MediationManager
from ..messages.mediate_grant import MediationGrant
from ..models.mediation_record import MediationRecord


class MediationGrantHandler(BaseHandler):
    """Handler for mediate-grant message."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle mediate-grant message."""
        self._logger.debug("%s called with context %s", self.__class__.__name__, context)
        assert isinstance(context.message, MediationGrant)

        if not context.connection_ready:
            raise HandlerException("Received mediation grant from inactive connection")

        profile = context.profile
        mgr = MediationManager(profile)
        try:
            async with profile.session() as session:
                record = await MediationRecord.retrieve_by_connection_id(
                    session, context.connection_record.connection_id
                )
            await mgr.request_granted(record, context.message)

            # Multitenancy setup
            multitenant_mgr = profile.inject_or(BaseMultitenantManager)
            wallet_id = profile.settings.get("wallet.id")

            if multitenant_mgr and wallet_id:
                base_mediation_record = await multitenant_mgr.get_default_mediator()

                # If we have a base mediator and sub wallet mediator
                # we need to register the base mediator routing key at the
                # sub wallet mediator
                if base_mediation_record:
                    # Last routing key needs to be registered at the mediator
                    keylist_updates = await mgr.add_key(
                        base_mediation_record.routing_keys[-1]
                    )

                    await responder.send(
                        keylist_updates,
                        connection_id=context.connection_record.connection_id,
                    )

            # Set to default if metadata set on connection to do so
            async with profile.session() as session:
                mediationRecord = await context.connection_record.metadata_get(
                    session, MediationManager.SET_TO_DEFAULT_ON_GRANTED
                )

            if mediationRecord:
                await mgr.set_default_mediator(record)

        except StorageNotFoundError as err:
            raise HandlerException(
                "Received mediation grant from connection from which mediation "
                "has not been requested."
            ) from err



TEST_LD_DOCUMENT_SIGNED_ED25519 = {
    "@context": ["http://schema.org/", "https://w3id.org/security/bbs/v1"],
    "@type": "Person",
    "firstName": "Jane",
    "lastName": "Does",
    "jobTitle": "Professor",
    "telephone": "(425) 123-4567",
    "email": "jane.doe@example.com",
    "proof": {
        "type": "Ed25519Signature2018",
        "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "created": "2021-04-26T20:21:49.045302",
        "proofPurpose": "assertionMethod",
        "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..UnUUc0S8ww-3fhX40IFfz9Ud_fqHurW4b4uovzBK3vQ8sVG-TMcXGNpC_v9QGZmWiOBOLNdOQOXUzLj2G88VCw",
    },
}



MULTIKEY_V1 = {
    "@context": {
        "id": "@id",
        "type": "@type",
        "@protected": True,
        "Multikey": {
            "@id": "https://w3id.org/security#Multikey",
            "@context": {
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "controller": {
                    "@id": "https://w3id.org/security#controller",
                    "@type": "@id",
                },
                "revoked": {
                    "@id": "https://w3id.org/security#revoked",
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                },
                "expires": {
                    "@id": "https://w3id.org/security#expiration",
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                },
                "publicKeyMultibase": {
                    "@id": "https://w3id.org/security#publicKeyMultibase",
                    "@type": "https://w3id.org/security#multibase",
                },
                "secretKeyMultibase": {
                    "@id": "https://w3id.org/security#secretKeyMultibase",
                    "@type": "https://w3id.org/security#multibase",
                },
            },
        },
    }
}



"""Library version information."""

RECORD_TYPE_ACAPY_STORAGE_TYPE = "acapy_storage_type"
RECORD_TYPE_ACAPY_UPGRADING = "acapy_upgrading"

STORAGE_TYPE_VALUE_ANONCREDS = "askar-anoncreds"
STORAGE_TYPE_VALUE_ASKAR = "askar"



"""Functions for upgrading records to anoncreds."""

import asyncio
import json
import logging
from typing import Optional

from anoncreds import (
    CredentialDefinition,
    CredentialDefinitionPrivate,
    KeyCorrectnessProof,
    RevocationRegistryDefinitionPrivate,
    Schema,
)
from aries_askar import AskarError
from indy_credx import LinkSecret

from ..anoncreds.issuer import (
    CATEGORY_CRED_DEF,
    CATEGORY_CRED_DEF_KEY_PROOF,
    CATEGORY_CRED_DEF_PRIVATE,
    CATEGORY_SCHEMA,
)
from ..anoncreds.models.credential_definition import CredDef, CredDefState
from ..anoncreds.models.revocation import (
    RevList,
    RevListState,
    RevRegDef,
    RevRegDefState,
    RevRegDefValue,
)
from ..anoncreds.models.schema import SchemaState
from ..anoncreds.revocation import (
    CATEGORY_REV_LIST,
    CATEGORY_REV_REG_DEF,
    CATEGORY_REV_REG_DEF_PRIVATE,
)
from ..cache.base import BaseCache
from ..core.profile import Profile
from ..indy.credx.holder import CATEGORY_LINK_SECRET, IndyCredxHolder
from ..ledger.multiple_ledger.ledger_requests_executor import (
    GET_CRED_DEF,
    GET_SCHEMA,
    IndyLedgerRequestsExecutor,
)
from ..messaging.credential_definitions.util import CRED_DEF_SENT_RECORD_TYPE
from ..messaging.schemas.util import SCHEMA_SENT_RECORD_TYPE
from ..multitenant.base import BaseMultitenantManager
from ..revocation.models.issuer_cred_rev_record import IssuerCredRevRecord
from ..revocation.models.issuer_rev_reg_record import IssuerRevRegRecord
from ..storage.base import BaseStorage
from ..storage.error import StorageNotFoundError
from ..storage.record import StorageRecord
from ..storage.type import (
    RECORD_TYPE_ACAPY_STORAGE_TYPE,
    RECORD_TYPE_ACAPY_UPGRADING,
    STORAGE_TYPE_VALUE_ANONCREDS,
)
from .singletons import IsAnonCredsSingleton, UpgradeInProgressSingleton

LOGGER = logging.getLogger(__name__)

UPGRADING_RECORD_IN_PROGRESS = "anoncreds_in_progress"
UPGRADING_RECORD_FINISHED = "anoncreds_finished"

# Number of times to retry upgrading records
max_retries = 5


class SchemaUpgradeObj:
    """Schema upgrade object."""

    def __init__(
        self,
        schema_id: str,
        schema: Schema,
        name: str,
        version: str,
        issuer_id: str,
        old_record_id: str,
    ):
        """Initialize schema upgrade object."""
        self.schema_id = schema_id
        self.schema = schema
        self.name = name
        self.version = version
        self.issuer_id = issuer_id
        self.old_record_id = old_record_id


class CredDefUpgradeObj:
    """Cred def upgrade object."""

    def __init__(
        self,
        cred_def_id: str,
        cred_def: CredentialDefinition,
        cred_def_private: CredentialDefinitionPrivate,
        key_proof: KeyCorrectnessProof,
        revocation: Optional[bool] = None,
        askar_cred_def: Optional[any] = None,
        max_cred_num: Optional[int] = None,
    ):
        """Initialize cred def upgrade object."""
        self.cred_def_id = cred_def_id
        self.cred_def = cred_def
        self.cred_def_private = cred_def_private
        self.key_proof = key_proof
        self.revocation = revocation
        self.askar_cred_def = askar_cred_def
        self.max_cred_num = max_cred_num


class RevRegDefUpgradeObj:
    """Rev reg def upgrade object."""

    def __init__(
        self,
        rev_reg_def_id: str,
        rev_reg_def: RevRegDef,
        rev_reg_def_private: RevocationRegistryDefinitionPrivate,
        active: bool = False,
    ):
        """Initialize rev reg def upgrade object."""
        self.rev_reg_def_id = rev_reg_def_id
        self.rev_reg_def = rev_reg_def
        self.rev_reg_def_private = rev_reg_def_private
        self.active = active


class RevListUpgradeObj:
    """Rev entry upgrade object."""

    def __init__(
        self,
        rev_list: RevList,
        pending: list,
        rev_reg_def_id: str,
        cred_rev_records: list,
    ):
        """Initialize rev entry upgrade object."""
        self.rev_list = rev_list
        self.pending = pending
        self.rev_reg_def_id = rev_reg_def_id
        self.cred_rev_records = cred_rev_records


async def get_schema_upgrade_object(
    profile: Profile, schema_id: str, askar_schema
) -> SchemaUpgradeObj:
    """Get schema upgrade object."""

    async with profile.session() as session:
        schema_id = askar_schema.tags.get("schema_id")
        issuer_did = askar_schema.tags.get("schema_issuer_did")
        # Need to get schema from the ledger because the attribute names
        # are not stored in the wallet
        multitenant_mgr = session.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
        else:
            ledger_exec_inst = session.inject(IndyLedgerRequestsExecutor)

    _, ledger = await ledger_exec_inst.get_ledger_for_identifier(
        schema_id,
        txn_record_type=GET_SCHEMA,
    )
    async with ledger:
        schema_from_ledger = await ledger.get_schema(schema_id)

    return SchemaUpgradeObj(
        schema_id,
        Schema.create(
            schema_id,
            askar_schema.tags.get("schema_name"),
            issuer_did,
            schema_from_ledger["attrNames"],
        ),
        askar_schema.tags.get("schema_name"),
        askar_schema.tags.get("schema_version"),
        issuer_did,
        askar_schema.id,
    )


async def get_cred_def_upgrade_object(
    profile: Profile, askar_cred_def
) -> CredDefUpgradeObj:
    """Get cred def upgrade object."""
    cred_def_id = askar_cred_def.tags.get("cred_def_id")
    async with profile.session() as session:
        # Need to get cred_def from the ledger because the tag
        # is not stored in the wallet and don't know wether it supports revocation
        multitenant_mgr = session.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
        else:
            ledger_exec_inst = session.inject(IndyLedgerRequestsExecutor)
    _, ledger = await ledger_exec_inst.get_ledger_for_identifier(
        cred_def_id,
        txn_record_type=GET_CRED_DEF,
    )
    async with ledger:
        cred_def_from_ledger = await ledger.get_credential_definition(cred_def_id)

    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        askar_cred_def_private = await storage.get_record(
            CATEGORY_CRED_DEF_PRIVATE, cred_def_id
        )
        askar_cred_def_key_proof = await storage.get_record(
            CATEGORY_CRED_DEF_KEY_PROOF, cred_def_id
        )

    cred_def = CredDef(
        issuer_id=askar_cred_def.tags.get("issuer_did"),
        schema_id=askar_cred_def.tags.get("schema_id"),
        tag=cred_def_from_ledger["tag"],
        type=cred_def_from_ledger["type"],
        value=cred_def_from_ledger["value"],
    )

    return CredDefUpgradeObj(
        cred_def_id,
        cred_def,
        askar_cred_def_private.value,
        askar_cred_def_key_proof.value,
        cred_def_from_ledger["value"].get("revocation", None),
        askar_cred_def=askar_cred_def,
    )


async def get_rev_reg_def_upgrade_object(
    profile: Profile,
    cred_def_upgrade_obj: CredDefUpgradeObj,
    askar_issuer_rev_reg_def,
    is_active: bool,
) -> RevRegDefUpgradeObj:
    """Get rev reg def upgrade object."""
    rev_reg_def_id = askar_issuer_rev_reg_def.tags.get("revoc_reg_id")

    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        askar_reg_rev_def_private = await storage.get_record(
            CATEGORY_REV_REG_DEF_PRIVATE, rev_reg_def_id
        )

    revoc_reg_def_values = json.loads(askar_issuer_rev_reg_def.value)

    reg_def_value = RevRegDefValue(
        revoc_reg_def_values["revoc_reg_def"]["value"]["publicKeys"],
        revoc_reg_def_values["revoc_reg_def"]["value"]["maxCredNum"],
        revoc_reg_def_values["revoc_reg_def"]["value"]["tailsLocation"],
        revoc_reg_def_values["revoc_reg_def"]["value"]["tailsHash"],
    )

    rev_reg_def = RevRegDef(
        issuer_id=askar_issuer_rev_reg_def.tags.get("issuer_did"),
        cred_def_id=cred_def_upgrade_obj.cred_def_id,
        tag=revoc_reg_def_values["tag"],
        type=revoc_reg_def_values["revoc_def_type"],
        value=reg_def_value,
    )

    return RevRegDefUpgradeObj(
        rev_reg_def_id, rev_reg_def, askar_reg_rev_def_private.value, is_active
    )


async def get_rev_list_upgrade_object(
    profile: Profile, rev_reg_def_upgrade_obj: RevRegDefUpgradeObj
) -> RevListUpgradeObj:
    """Get revocation entry upgrade object."""
    rev_reg = rev_reg_def_upgrade_obj.rev_reg_def
    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        askar_cred_rev_records = await storage.find_all_records(
            IssuerCredRevRecord.RECORD_TYPE,
            {"rev_reg_id": rev_reg_def_upgrade_obj.rev_reg_def_id},
        )

    revocation_list = [0] * rev_reg.value.max_cred_num
    for askar_cred_rev_record in askar_cred_rev_records:
        if askar_cred_rev_record.tags.get("state") == "revoked":
            revocation_list[int(askar_cred_rev_record.tags.get("cred_rev_id")) - 1] = 1

    rev_list = RevList(
        issuer_id=rev_reg.issuer_id,
        rev_reg_def_id=rev_reg_def_upgrade_obj.rev_reg_def_id,
        revocation_list=revocation_list,
        current_accumulator=json.loads(
            rev_reg_def_upgrade_obj.askar_issuer_rev_reg_def.value
        )["revoc_reg_entry"]["value"]["accum"],
    )

    return RevListUpgradeObj(
        rev_list,
        json.loads(rev_reg_def_upgrade_obj.askar_issuer_rev_reg_def.value)["pending_pub"],
        rev_reg_def_upgrade_obj.rev_reg_def_id,
        askar_cred_rev_records,
    )


async def upgrade_and_delete_schema_records(
    txn, schema_upgrade_obj: SchemaUpgradeObj
) -> None:
    """Upgrade and delete schema records."""
    schema_anoncreds = schema_upgrade_obj.schema
    await txn.handle.remove("schema_sent", schema_upgrade_obj.old_record_id)
    await txn.handle.replace(
        CATEGORY_SCHEMA,
        schema_upgrade_obj.schema_id,
        schema_anoncreds.to_json(),
        {
            "name": schema_upgrade_obj.name,
            "version": schema_upgrade_obj.version,
            "issuer_id": schema_upgrade_obj.issuer_id,
            "state": SchemaState.STATE_FINISHED,
        },
    )


async def upgrade_and_delete_cred_def_records(
    txn, anoncreds_schema, cred_def_upgrade_obj: CredDefUpgradeObj
) -> None:
    """Upgrade and delete cred def records."""
    cred_def_id = cred_def_upgrade_obj.cred_def_id
    anoncreds_schema = anoncreds_schema.to_dict()
    askar_cred_def = cred_def_upgrade_obj.askar_cred_def
    await txn.handle.remove("cred_def_sent", askar_cred_def.id)
    await txn.handle.replace(
        CATEGORY_CRED_DEF,
        cred_def_id,
        cred_def_upgrade_obj.cred_def.to_json(),
        tags={
            "schema_id": askar_cred_def.tags.get("schema_id"),
            "schema_issuer_id": anoncreds_schema["issuerId"],
            "issuer_id": askar_cred_def.tags.get("issuer_did"),
            "schema_name": anoncreds_schema["name"],
            "schema_version": anoncreds_schema["version"],
            "state": CredDefState.STATE_FINISHED,
            "epoch": askar_cred_def.tags.get("epoch"),
            # TODO We need to keep track of these but tags probably
            # isn't ideal. This suggests that a full record object
            # is necessary for non-private values
            "support_revocation": json.dumps(cred_def_upgrade_obj.revocation),
            "max_cred_num": str(cred_def_upgrade_obj.max_cred_num or 0),
        },
    )
    await txn.handle.replace(
        CATEGORY_CRED_DEF_PRIVATE,
        cred_def_id,
        CredentialDefinitionPrivate.load(
            cred_def_upgrade_obj.cred_def_private
        ).to_json_buffer(),
    )
    await txn.handle.replace(
        CATEGORY_CRED_DEF_KEY_PROOF,
        cred_def_id,
        KeyCorrectnessProof.load(cred_def_upgrade_obj.key_proof).to_json_buffer(),
    )


rev_reg_states_mapping = {
    "init": RevRegDefState.STATE_WAIT,
    "generated": RevRegDefState.STATE_ACTION,
    "posted": RevRegDefState.STATE_FINISHED,
    "active": RevRegDefState.STATE_FINISHED,
    "full": RevRegDefState.STATE_FULL,
    "decommissioned": RevRegDefState.STATE_DECOMMISSIONED,
}


async def upgrade_and_delete_rev_reg_def_records(
    txn, rev_reg_def_upgrade_obj: RevRegDefUpgradeObj
) -> None:
    """Upgrade and delete rev reg def records."""
    rev_reg_def_id = rev_reg_def_upgrade_obj.rev_reg_def_id
    askar_issuer_rev_reg_def = rev_reg_def_upgrade_obj.askar_issuer_rev_reg_def
    await txn.handle.remove(IssuerRevRegRecord.RECORD_TYPE, askar_issuer_rev_reg_def.id)
    await txn.handle.replace(
        CATEGORY_REV_REG_DEF,
        rev_reg_def_id,
        rev_reg_def_upgrade_obj.rev_reg_def.to_json(),
        tags={
            "cred_def_id": rev_reg_def_upgrade_obj.rev_reg_def.cred_def_id,
            "issuer_id": askar_issuer_rev_reg_def.tags.get("issuer_did"),
            "state": rev_reg_states_mapping[askar_issuer_rev_reg_def.tags.get("state")],
            "active": json.dumps(rev_reg_def_upgrade_obj.active),
        },
    )
    await txn.handle.replace(
        CATEGORY_REV_REG_DEF_PRIVATE,
        rev_reg_def_id,
        RevocationRegistryDefinitionPrivate.load(
            rev_reg_def_upgrade_obj.rev_reg_def_private
        ).to_json_buffer(),
    )


async def upgrade_and_delete_rev_entry_records(
    txn, rev_list_upgrade_obj: RevListUpgradeObj
) -> None:
    """Upgrade and delete revocation entry records."""
    next_index = 0
    for cred_rev_record in rev_list_upgrade_obj.cred_rev_records:
        if int(cred_rev_record.tags.get("cred_rev_id")) > next_index:
            next_index = int(cred_rev_record.tags.get("cred_rev_id"))
        await txn.handle.remove(IssuerCredRevRecord.RECORD_TYPE, cred_rev_record.id)

    await txn.handle.insert(
        CATEGORY_REV_LIST,
        rev_list_upgrade_obj.rev_reg_def_id,
        value_json={
            "rev_list": rev_list_upgrade_obj.rev_list.serialize(),
            "pending": rev_list_upgrade_obj.pending,
            "next_index": next_index + 1,
        },
        tags={
            "state": RevListState.STATE_FINISHED,
            "pending": json.dumps(rev_list_upgrade_obj.pending is not None),
        },
    )


async def upgrade_all_records_with_transaction(
    txn: any,
    schema_upgrade_objs: list[SchemaUpgradeObj],
    cred_def_upgrade_objs: list[CredDefUpgradeObj],
    rev_reg_def_upgrade_objs: list[RevRegDefUpgradeObj],
    rev_list_upgrade_objs: list[RevListUpgradeObj],
    link_secret: Optional[LinkSecret] = None,
) -> None:
    """Upgrade all objects with transaction."""
    for schema_upgrade_obj in schema_upgrade_objs:
        await upgrade_and_delete_schema_records(txn, schema_upgrade_obj)
    for cred_def_upgrade_obj in cred_def_upgrade_objs:
        await upgrade_and_delete_cred_def_records(
            txn, schema_upgrade_obj.schema, cred_def_upgrade_obj
        )
    for rev_reg_def_upgrade_obj in rev_reg_def_upgrade_objs:
        await upgrade_and_delete_rev_reg_def_records(txn, rev_reg_def_upgrade_obj)
    for rev_list_upgrade_obj in rev_list_upgrade_objs:
        await upgrade_and_delete_rev_entry_records(txn, rev_list_upgrade_obj)

    if link_secret:
        await txn.handle.replace(
            CATEGORY_LINK_SECRET,
            IndyCredxHolder.LINK_SECRET_ID,
            link_secret.to_dict()["value"]["ms"].encode("ascii"),
        )

    await txn.commit()


async def get_rev_reg_def_upgrade_objs(
    profile: Profile,
    cred_def_upgrade_obj: CredDefUpgradeObj,
    rev_list_upgrade_objs: list[RevListUpgradeObj],
) -> list[RevRegDefUpgradeObj]:
    """Get rev reg def upgrade objects."""

    rev_reg_def_upgrade_objs = []
    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        # Must be sorted to find the active rev reg def
        askar_issuer_rev_reg_def_records = sorted(
            await storage.find_all_records(
                IssuerRevRegRecord.RECORD_TYPE,
                {"cred_def_id": cred_def_upgrade_obj.cred_def_id},
            ),
            key=lambda x: json.loads(x.value)["created_at"],
        )
    found_active = False
    is_active = False
    for askar_issuer_rev_reg_def in askar_issuer_rev_reg_def_records:
        # active rev reg def is the oldest non-full and active rev reg def
        if (
            not found_active
            and askar_issuer_rev_reg_def.tags.get("state") != "full"
            and askar_issuer_rev_reg_def.tags.get("state") == "active"
        ):
            found_active = True
            is_active = True

        rev_reg_def_upgrade_obj = await get_rev_reg_def_upgrade_object(
            profile,
            cred_def_upgrade_obj,
            askar_issuer_rev_reg_def,
            is_active,
        )
        is_active = False
        rev_reg_def_upgrade_obj.askar_issuer_rev_reg_def = askar_issuer_rev_reg_def

        rev_reg_def_upgrade_objs.append(rev_reg_def_upgrade_obj)

        # add the revocation list upgrade object from reg def upgrade object
        rev_list_upgrade_objs.append(
            await get_rev_list_upgrade_object(profile, rev_reg_def_upgrade_obj)
        )
    return rev_reg_def_upgrade_objs


async def convert_records_to_anoncreds(profile) -> None:
    """Convert and delete old askar records."""
    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        askar_schema_records = await storage.find_all_records(SCHEMA_SENT_RECORD_TYPE)

        schema_upgrade_objs = []
        cred_def_upgrade_objs = []
        rev_reg_def_upgrade_objs = []
        rev_list_upgrade_objs = []

        # Schemas
        for askar_schema in askar_schema_records:
            schema_upgrade_objs.append(
                await get_schema_upgrade_object(profile, askar_schema.id, askar_schema)
            )

        # CredDefs and Revocation Objects
        askar_cred_def_records = await storage.find_all_records(
            CRED_DEF_SENT_RECORD_TYPE, {}
        )
        for askar_cred_def in askar_cred_def_records:
            cred_def_upgrade_obj = await get_cred_def_upgrade_object(
                profile, askar_cred_def
            )
            rev_reg_def_upgrade_objs = await get_rev_reg_def_upgrade_objs(
                profile, cred_def_upgrade_obj, rev_list_upgrade_objs
            )
            # update the cred_def with the max_cred_num from first rev_reg_def
            if rev_reg_def_upgrade_objs:
                cred_def_upgrade_obj.max_cred_num = rev_reg_def_upgrade_objs[
                    0
                ].rev_reg_def.value.max_cred_num
            cred_def_upgrade_objs.append(cred_def_upgrade_obj)

        # Link secret
        link_secret_record = None
        try:
            link_secret_record = await session.handle.fetch(
                CATEGORY_LINK_SECRET, IndyCredxHolder.LINK_SECRET_ID
            )
        except AskarError:
            pass

        link_secret = None
        if link_secret_record:
            link_secret = LinkSecret.load(link_secret_record.raw_value)

        async with profile.transaction() as txn:
            try:
                await upgrade_all_records_with_transaction(
                    txn,
                    schema_upgrade_objs,
                    cred_def_upgrade_objs,
                    rev_reg_def_upgrade_objs,
                    rev_list_upgrade_objs,
                    link_secret,
                )
            except Exception as e:
                await txn.rollback()
                raise e


async def retry_converting_records(
    profile: Profile, upgrading_record: StorageRecord, retry: int, is_subwallet=False
) -> None:
    """Retry converting records to anoncreds."""

    async def fail_upgrade():
        async with profile.session() as session:
            storage = session.inject(BaseStorage)
            await storage.delete_record(upgrading_record)

    try:
        await convert_records_to_anoncreds(profile)
        await finish_upgrade_by_updating_profile_or_shutting_down(profile, is_subwallet)
        LOGGER.info(f"Upgrade complete via retry for wallet: {profile.name}")
    except Exception as e:
        LOGGER.error(f"Error when upgrading records for wallet {profile.name} : {e} ")
        if retry < max_retries:
            LOGGER.info(f"Retry attempt {retry + 1} to upgrade wallet {profile.name}")
            await asyncio.sleep(1)
            await retry_converting_records(
                profile, upgrading_record, retry + 1, is_subwallet
            )
        else:
            LOGGER.error(
                f"""Failed to upgrade wallet: {profile.name} after 5 retries. 
                Try fixing any connection issues and re-running the update"""
            )
            await fail_upgrade()


async def upgrade_wallet_to_anoncreds_if_requested(
    profile: Profile, is_subwallet=False
) -> None:
    """Get upgrading record and attempt to upgrade wallet to anoncreds."""
    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        try:
            upgrading_record = await storage.find_record(RECORD_TYPE_ACAPY_UPGRADING, {})
            if upgrading_record.value == UPGRADING_RECORD_FINISHED:
                IsAnonCredsSingleton().set_wallet(profile.name)
                return
        except StorageNotFoundError:
            return

        try:
            LOGGER.info("Upgrade in process for wallet: %s", profile.name)
            await convert_records_to_anoncreds(profile)
            await finish_upgrade_by_updating_profile_or_shutting_down(
                profile, is_subwallet
            )
        except Exception as e:
            LOGGER.error(f"Error when upgrading wallet {profile.name} : {e} ")
            await retry_converting_records(profile, upgrading_record, 0, is_subwallet)


async def finish_upgrade(profile: Profile):
    """Finish record by setting records and caches."""
    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        try:
            storage_type_record = await storage.find_record(
                type_filter=RECORD_TYPE_ACAPY_STORAGE_TYPE, tag_query={}
            )
            await storage.update_record(
                storage_type_record, STORAGE_TYPE_VALUE_ANONCREDS, {}
            )
        # This should only happen for subwallets
        except StorageNotFoundError:
            await storage.add_record(
                StorageRecord(
                    RECORD_TYPE_ACAPY_STORAGE_TYPE,
                    STORAGE_TYPE_VALUE_ANONCREDS,
                )
            )
    await finish_upgrading_record(profile)
    IsAnonCredsSingleton().set_wallet(profile.name)
    UpgradeInProgressSingleton().remove_wallet(profile.name)


async def finish_upgrading_record(profile: Profile):
    """Update upgrading record to finished."""
    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        try:
            upgrading_record = await storage.find_record(
                RECORD_TYPE_ACAPY_UPGRADING, tag_query={}
            )
            await storage.update_record(upgrading_record, UPGRADING_RECORD_FINISHED, {})
        except StorageNotFoundError:
            return


async def upgrade_subwallet(profile: Profile) -> None:
    """Upgrade subwallet to anoncreds."""
    async with profile.session() as session:
        multitenant_mgr = session.inject_or(BaseMultitenantManager)
        wallet_id = profile.settings.get("wallet.id")
        cache = profile.inject_or(BaseCache)
        await cache.flush()
        settings = {"wallet.type": STORAGE_TYPE_VALUE_ANONCREDS}
        await multitenant_mgr.update_wallet(wallet_id, settings)


async def finish_upgrade_by_updating_profile_or_shutting_down(
    profile: Profile, is_subwallet=False
):
    """Upgrade wallet to anoncreds and set storage type."""
    if is_subwallet:
        await upgrade_subwallet(profile)
        await finish_upgrade(profile)
        LOGGER.info(
            "Upgrade of subwallet %s has completed. Profile is now askar-anoncreds",
            profile.settings.get("wallet.name"),
        )
    else:
        await finish_upgrade(profile)
        LOGGER.info(
            "Upgrade of base wallet %s to anoncreds has completed. Shutting down agent.",
            profile.settings.get("wallet.name"),
        )
        asyncio.get_event_loop().stop()


async def check_upgrade_completion_loop(profile: Profile, is_subwallet=False):
    """Check if upgrading is complete."""
    async with profile.session() as session:
        while True:
            storage = session.inject(BaseStorage)
            LOGGER.debug("Checking upgrade completion for wallet: %s", profile.name)
            try:
                upgrading_record = await storage.find_record(
                    RECORD_TYPE_ACAPY_UPGRADING, tag_query={}
                )
                if upgrading_record.value == UPGRADING_RECORD_FINISHED:
                    IsAnonCredsSingleton().set_wallet(profile.name)
                    UpgradeInProgressSingleton().remove_wallet(profile.name)
                    if is_subwallet:
                        await upgrade_subwallet(profile)
                        await finish_upgrade(profile)
                        LOGGER.info(
                            "Upgrade of subwallet %s has completed. "
                            "Profile is now askar-anoncreds",
                            profile.settings.get("wallet.name"),
                        )
                        return
                    LOGGER.info(
                        "Upgrade complete for wallet: %s, shutting down agent.",
                        profile.name,
                    )
                    # Shut down agent if base wallet
                    asyncio.get_event_loop().stop()
            except StorageNotFoundError:
                # If the record is not found, the upgrade failed
                return

            await asyncio.sleep(1)



"""Test JwkDIDResolver."""

import pytest
import pytest_asyncio

from ....core.profile import Profile
from ....utils.testing import create_test_profile
from ...base import DIDMethodNotSupported
from ..jwk import JwkDIDResolver

# pylint: disable=W0621
TEST_DIDS = [
    "did:jwk:eyJjcnYiOiJQLTI1NiIsImt0eSI6IkVDIiwieCI6ImFjYklRaXVNczNpOF91c3pFakoydHBUdFJNNEVVM3l6OTFQSDZDZEgyVjAiLCJ5IjoiX0tjeUxqOXZXTXB0bm1LdG00NkdxRHo4d2Y3NEk1TEtncmwyR3pIM25TRSJ9",
    "did:jwk:eyJrdHkiOiJPS1AiLCJjcnYiOiJYMjU1MTkiLCJ1c2UiOiJlbmMiLCJ4IjoiM3A3YmZYdDl3YlRUVzJIQzdPUTFOei1EUThoYmVHZE5yZngtRkctSUswOCJ9",
]

TEST_DID_INVALID = "did:key:z1MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th"


@pytest.fixture
def resolver():
    """Resolver fixture."""
    yield JwkDIDResolver()


@pytest_asyncio.fixture
async def profile():
    """Profile fixture."""
    profile = await create_test_profile()
    yield profile


@pytest.mark.asyncio
@pytest.mark.parametrize("did", TEST_DIDS)
async def test_supported_did_regex(profile, resolver: JwkDIDResolver, did: str):
    """Test the supported_did_regex."""
    assert await resolver.supports(
        profile,
        did,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("did", TEST_DIDS)
async def test_resolve(resolver: JwkDIDResolver, profile: Profile, did: str):
    """Test resolve method."""
    assert await resolver.resolve(profile, did)


@pytest.mark.asyncio
async def test_resolve_x_did_not_found(resolver: JwkDIDResolver, profile: Profile):
    """Test resolve method when no did is found."""
    with pytest.raises(DIDMethodNotSupported):
        await resolver.resolve(profile, TEST_DID_INVALID)



"""Abstract base classes for non-secrets storage."""

from abc import ABC, abstractmethod
from typing import Mapping, Optional, Sequence

from .error import StorageDuplicateError, StorageError, StorageNotFoundError
from .record import StorageRecord

DEFAULT_PAGE_SIZE = 100
MAXIMUM_PAGE_SIZE = 10000


def validate_record(record: StorageRecord, *, delete=False):
    """Ensure that a record is ready to be saved/updated/deleted."""
    if not record:
        raise StorageError("No record provided")
    if not record.id:
        raise StorageError("Record has no ID")
    if not record.type:
        raise StorageError("Record has no type")
    if not record.value and not delete:
        raise StorageError("Record must have a non-empty value")


class BaseStorage(ABC):
    """Abstract stored records interface."""

    @abstractmethod
    async def add_record(self, record: StorageRecord):
        """Add a new record to the store.

        Args:
            record: `StorageRecord` to be stored

        """

    @abstractmethod
    async def get_record(
        self, record_type: str, record_id: str, options: Optional[Mapping] = None
    ) -> StorageRecord:
        """Fetch a record from the store by type and ID.

        Args:
            record_type: The record type
            record_id: The record id
            options: A dictionary of backend-specific options

        Returns:
            A `StorageRecord` instance

        """

    @abstractmethod
    async def update_record(self, record: StorageRecord, value: str, tags: Mapping):
        """Update an existing stored record's value and tags.

        Args:
            record: `StorageRecord` to update
            value: The new value
            tags: The new tags

        """

    @abstractmethod
    async def delete_record(self, record: StorageRecord):
        """Delete an existing record.

        Args:
            record: `StorageRecord` to delete

        """

    async def find_record(
        self,
        type_filter: str,
        tag_query: Optional[Mapping] = None,
        options: Optional[Mapping] = None,
    ) -> StorageRecord:
        """Find a record using a unique tag filter.

        Args:
            type_filter: Filter string
            tag_query: Tags to query
            options: Dictionary of backend-specific options
        """
        scan = self.search_records(type_filter, tag_query, options)
        results = await scan.fetch(2)
        await scan.close()
        if not results:
            raise StorageNotFoundError("Record not found")
        if len(results) > 1:
            raise StorageDuplicateError("Duplicate records found")
        return results[0]

    @abstractmethod
    async def find_paginated_records(
        self,
        type_filter: str,
        tag_query: Optional[Mapping] = None,
        limit: int = DEFAULT_PAGE_SIZE,
        offset: int = 0,
        order_by: Optional[str] = None,
        descending: bool = False,
    ) -> Sequence[StorageRecord]:
        """Retrieve a page of records matching a particular type filter and tag query.

        Args:
            type_filter: The type of records to filter by
            tag_query: An optional dictionary of tag filter clauses
            limit: The maximum number of records to retrieve
            offset: The offset to start retrieving records from
            order_by: An optional field by which to order the records.
            descending: Whether to order the records in descending order.

        Returns:
            A sequence of StorageRecord matching the filter and query parameters.
        """

    @abstractmethod
    async def find_all_records(
        self,
        type_filter: str,
        tag_query: Optional[Mapping] = None,
        order_by: Optional[str] = None,
        descending: bool = False,
        options: Optional[Mapping] = None,
    ) -> Sequence[StorageRecord]:
        """Retrieve all records matching a particular type filter and tag query.

        Args:
            type_filter: The type of records to filter by.
            tag_query: An optional dictionary of tag filter clauses.
            order_by: An optional field by which to order the records.
            descending: Whether to order the records in descending order.
            options: Additional options for the query.
        """

    @abstractmethod
    async def delete_all_records(
        self,
        type_filter: str,
        tag_query: Optional[Mapping] = None,
    ) -> None:
        """Remove all records matching a particular type filter and tag query.

        Args:
            type_filter: The type of records to filter by.
            tag_query: An optional dictionary of tag filter clauses.
        """


class BaseStorageSearch(ABC):
    """Abstract stored records search interface."""

    @abstractmethod
    def search_records(
        self,
        type_filter: str,
        tag_query: Optional[Mapping] = None,
        page_size: Optional[int] = None,
        options: Optional[Mapping] = None,
    ) -> "BaseStorageSearchSession":
        """Create a new record query.

        Args:
            type_filter: Filter string
            tag_query: Tags to query
            page_size: Page size
            options: Dictionary of backend-specific options

        Returns:
            An instance of `BaseStorageSearchSession`

        """

    def __repr__(self) -> str:
        """Human readable representation of a `BaseStorage` implementation."""
        return "<{}>".format(self.__class__.__name__)


class BaseStorageSearchSession(ABC):
    """Abstract stored records search session interface."""

    @abstractmethod
    async def fetch(self, max_count: Optional[int] = None) -> Sequence[StorageRecord]:
        """Fetch the next list of results from the store.

        Args:
            max_count: Max number of records to return. If not provided,
              defaults to the backend's preferred page size

        Returns:
            A list of `StorageRecord` instances

        """

    async def close(self):
        """Dispose of the search query."""

    def __aiter__(self):
        """Async iterator magic method."""
        return IterSearch(self)

    def __repr__(self) -> str:
        """Human readable representation of this instance."""
        return "<{}>".format(self.__class__.__name__)


class IterSearch:
    """A generic record search async iterator."""

    def __init__(self, search: BaseStorageSearchSession, page_size: Optional[int] = None):
        """Instantiate a new `IterSearch` instance."""
        self._buffer = None
        self._page_size = page_size
        self._search = search

    async def __anext__(self):
        """Async iterator magic method."""
        if not self._buffer:
            self._buffer = await self._search.fetch(self._page_size) or []
        try:
            return self._buffer.pop(0)
        except IndexError:
            raise StopAsyncIteration



"""Indy DID Resolver.

Resolution is performed using the IndyLedger class.
"""

import logging
from typing import Optional, Pattern, Sequence, Text

from pydid import DID, DIDDocumentBuilder
from pydid.verification_method import Ed25519VerificationKey2018, VerificationMethod

from ...config.injection_context import InjectionContext
from ...core.profile import Profile
from ...did.did_key import DIDKey
from ...ledger.endpoint_type import EndpointType
from ...ledger.error import LedgerError
from ...ledger.multiple_ledger.ledger_requests_executor import (
    GET_KEY_FOR_DID,
    IndyLedgerRequestsExecutor,
)
from ...messaging.valid import IndyDID
from ...multitenant.base import BaseMultitenantManager
from ...wallet.key_type import ED25519
from ..base import BaseDIDResolver, DIDNotFound, ResolverError, ResolverType

LOGGER = logging.getLogger(__name__)


class NoIndyLedger(ResolverError):
    """Raised when there is no Indy ledger instance configured."""


def _routing_keys_as_did_key_urls(routing_keys: Sequence[str]) -> Sequence[str]:
    """Convert raw base58 keys to did:key values.

    If a did:key is passed in, convert to a did:key URL.
    """

    did_key_urls = []
    for routing_key in routing_keys:
        if not routing_key.startswith("did:key:"):
            did_key_urls.append(DIDKey.from_public_key_b58(routing_key, ED25519).key_id)
        else:
            if "#" not in routing_key:
                did_key_urls.append(
                    f"{routing_key}#{DIDKey.from_did(routing_key).fingerprint}"
                )
            else:
                return routing_keys
    return did_key_urls


class IndyDIDResolver(BaseDIDResolver):
    """Indy DID Resolver."""

    SERVICE_TYPE_DID_COMMUNICATION = "did-communication"
    SERVICE_TYPE_DIDCOMM = "DIDComm"
    SERVICE_TYPE_ENDPOINT = "endpoint"
    CONTEXT_DIDCOMM_V2 = "https://didcomm.org/messaging/contexts/v2"

    def __init__(self):
        """Initialize Indy Resolver."""
        super().__init__(ResolverType.NATIVE)

    async def setup(self, context: InjectionContext):
        """Perform required setup for Indy DID resolution."""

    @property
    def supported_did_regex(self) -> Pattern:
        """Return supported_did_regex of Indy DID Resolver."""
        return IndyDID.PATTERN

    def process_endpoint_types(self, types):
        """Process endpoint types.

        Returns expected types, subset of expected types,
        or default types.
        """
        expected_types = ["endpoint", "did-communication", "DIDComm"]
        default_types = ["endpoint", "did-communication"]
        if len(types) <= 0:
            return default_types
        for type in types:
            if type not in expected_types:
                return default_types
        return types

    def add_services(
        self,
        builder: DIDDocumentBuilder,
        endpoints: Optional[dict],
        recipient_key: Optional[VerificationMethod] = None,
        service_accept: Optional[Sequence[Text]] = None,
    ):
        """Add services."""
        if not endpoints:
            return

        endpoint = endpoints.get("endpoint")
        routing_keys = endpoints.get("routingKeys", [])
        types = endpoints.get("types", [self.SERVICE_TYPE_DID_COMMUNICATION])

        other_endpoints = {
            key: endpoints[key]
            for key in ("profile", "linked_domains")
            if key in endpoints
        }

        if endpoint:
            processed_types = self.process_endpoint_types(types)

            if self.SERVICE_TYPE_ENDPOINT in processed_types:
                builder.service.add(
                    ident="endpoint",
                    service_endpoint=endpoint,
                    type_=self.SERVICE_TYPE_ENDPOINT,
                )

            if self.SERVICE_TYPE_DID_COMMUNICATION in processed_types:
                builder.service.add(
                    ident="did-communication",
                    type_=self.SERVICE_TYPE_DID_COMMUNICATION,
                    service_endpoint=endpoint,
                    priority=1,
                    routing_keys=_routing_keys_as_did_key_urls(routing_keys),
                    recipient_keys=[recipient_key.id],
                    accept=(
                        service_accept if service_accept else ["didcomm/aip2;env=rfc19"]
                    ),
                )

            if self.SERVICE_TYPE_DIDCOMM in types:
                builder.service.add(
                    ident="#didcomm-1",
                    type_=self.SERVICE_TYPE_DIDCOMM,
                    service_endpoint=endpoint,
                    recipient_keys=[recipient_key.id],
                    routing_keys=_routing_keys_as_did_key_urls(routing_keys),
                    # CHECKME
                    # accept=(service_accept if service_accept else ["didcomm/v2"]),
                    accept=["didcomm/v2"],
                )
                builder.context.append(self.CONTEXT_DIDCOMM_V2)
        else:
            LOGGER.warning("No endpoint for DID although endpoint attrib was resolvable")

        if other_endpoints:
            for type_, endpoint in other_endpoints.items():
                builder.service.add(
                    ident=type_,
                    type_=EndpointType.get(type_).w3c,
                    service_endpoint=endpoint,
                )

    async def _resolve(
        self,
        profile: Profile,
        did: str,
        service_accept: Optional[Sequence[Text]] = None,
    ) -> dict:
        """Resolve an indy DID."""
        multitenant_mgr = profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
        else:
            ledger_exec_inst = profile.inject(IndyLedgerRequestsExecutor)
        ledger = (
            await ledger_exec_inst.get_ledger_for_identifier(
                did,
                txn_record_type=GET_KEY_FOR_DID,
            )
        )[1]
        if not ledger:
            raise NoIndyLedger("No Indy ledger instance is configured.")

        try:
            async with ledger:
                recipient_key = await ledger.get_key_for_did(did)
                endpoints: Optional[dict] = await ledger.get_all_endpoints_for_did(did)
        except LedgerError as err:
            raise DIDNotFound(f"DID {did} could not be resolved") from err

        builder = DIDDocumentBuilder(DID(did))

        vmethod = builder.verification_method.add(
            Ed25519VerificationKey2018, ident="key-1", public_key_base58=recipient_key
        )
        builder.authentication.reference(vmethod.id)
        builder.assertion_method.reference(vmethod.id)
        self.add_services(builder, endpoints, vmethod, service_accept)

        result = builder.build()
        return result.serialize()



"""Test mediate deny message."""

from unittest import TestCase

from ...message_types import MEDIATE_DENY
from ..mediate_deny import MediationDeny, MediationDenySchema
from . import MessageTest


class TestMediateDeny(MessageTest, TestCase):
    """Test mediate deny message."""

    TYPE = MEDIATE_DENY
    CLASS = MediationDeny
    SCHEMA = MediationDenySchema
    VALUES = {}



"""Key type code."""

from typing import Optional


class KeyType:
    """Key Type class."""

    def __init__(
        self,
        key_type: str,
        multicodec_name: str,
        multicodec_prefix: bytes,
        jws_alg: Optional[str],
    ):
        """Construct key type."""
        self._type: str = key_type
        self._name: str = multicodec_name
        self._prefix: bytes = multicodec_prefix
        self._jws_alg: bytes = jws_alg

    @property
    def key_type(self) -> str:
        """Get Key type, type."""
        return self._type

    @property
    def multicodec_name(self) -> str:
        """Get key type multicodec name."""
        return self._name

    @property
    def multicodec_prefix(self) -> bytes:
        """Get key type multicodec prefix."""
        return self._prefix

    @property
    def jws_algorithm(self) -> Optional[str]:
        """Get key type JWS Algorithm (used in the JOSE header)."""
        return self._jws_alg


# NOTE: the py_multicodec library is outdated. We use hardcoded prefixes here
# until this PR gets released: https://github.com/multiformats/py-multicodec/pull/14
# multicodec is also not used now, but may be used again if py_multicodec is updated
ED25519: KeyType = KeyType("ed25519", "ed25519-pub", b"\xed\x01", "EdDSA")
X25519: KeyType = KeyType("x25519", "x25519-pub", b"\xec\x01", None)
P256: KeyType = KeyType("p256", "p256-pub", b"\x80\x24", "ES256")
BLS12381G1: KeyType = KeyType("bls12381g1", "bls12_381-g1-pub", b"\xea\x01", None)
BLS12381G2: KeyType = KeyType("bls12381g2", "bls12_381-g2-pub", b"\xeb\x01", None)
BLS12381G1G2: KeyType = KeyType("bls12381g1g2", "bls12_381-g1g2-pub", b"\xee\x01", None)


class KeyTypes:
    """KeyType class specifying key types with multicodec name."""

    def __init__(self) -> None:
        """Construct key type registry."""
        self._type_registry: dict[str, KeyType] = {
            ED25519.key_type: ED25519,
            X25519.key_type: X25519,
            P256.key_type: P256,
            BLS12381G1.key_type: BLS12381G1,
            BLS12381G2.key_type: BLS12381G2,
            BLS12381G1G2.key_type: BLS12381G1G2,
        }
        self._name_registry: dict[str, KeyType] = {
            ED25519.multicodec_name: ED25519,
            X25519.multicodec_name: X25519,
            P256.multicodec_name: P256,
            BLS12381G1.multicodec_name: BLS12381G1,
            BLS12381G2.multicodec_name: BLS12381G2,
            BLS12381G1G2.multicodec_name: BLS12381G1G2,
        }
        self._prefix_registry: dict[bytes, KeyType] = {
            ED25519.multicodec_prefix: ED25519,
            X25519.multicodec_prefix: X25519,
            P256.multicodec_prefix: P256,
            BLS12381G1.multicodec_prefix: BLS12381G1,
            BLS12381G2.multicodec_prefix: BLS12381G2,
            BLS12381G1G2.multicodec_prefix: BLS12381G1G2,
        }

    def register(self, key_type: KeyType):
        """Register a new key type."""
        self._type_registry[key_type.key_type] = key_type
        self._name_registry[key_type.multicodec_name] = key_type
        self._prefix_registry[key_type.multicodec_prefix] = key_type

    def from_multicodec_name(self, multicodec_name: str) -> Optional["KeyType"]:
        """Get KeyType instance based on multicodec name. Returns None if not found."""
        return self._name_registry.get(multicodec_name)

    def from_multicodec_prefix(self, multicodec_prefix: bytes) -> Optional["KeyType"]:
        """Get KeyType instance based on multicodec prefix. Returns None if not found."""
        return self._prefix_registry.get(multicodec_prefix)

    def from_prefixed_bytes(self, prefixed_bytes: bytes) -> Optional["KeyType"]:
        """Get KeyType instance based on prefix in bytes. Returns None if not found."""
        return next(
            (
                key_type
                for key_type in self._name_registry.values()
                if prefixed_bytes.startswith(key_type.multicodec_prefix)
            ),
            None,
        )

    def from_key_type(self, key_type: str) -> Optional["KeyType"]:
        """Get KeyType instance from the key type identifier."""
        return self._type_registry.get(key_type)



"""Quick and dirty fix to use as alternative to pyld downloader.

Allows keeping some context in local filesystem.
"""

import logging
import re
import string
import urllib.parse as urllib_parse
from importlib import resources
from typing import Dict, Optional

import requests
from pyld import jsonld
from pyld.jsonld import LINK_HEADER_REL, JsonLdError, parse_link_header

from ...version import __version__

logger = logging.getLogger(__name__)


def _load_jsonld_file(
    original_url, filename: str, resource_path: str = f"{__package__}.resources"
):
    """Load context from package.

    Given a URL and filename,
    load a context in the format used by pyld document loader.
    """
    return {
        "contentType": "application/ld+json",
        "contextUrl": None,
        "documentUrl": original_url,
        "document": (resources.files(resource_path) / filename).read_text(),
    }


class StaticCacheJsonLdDownloader:
    """Context downloader with filesystem static cache for common contexts."""

    CONTEXT_FILE_MAPPING = {
        "https://www.w3.org/2018/credentials/v1": "credentials_context.jsonld",
        "https://www.w3.org/ns/credentials/v2": "credentials_v2_context.jsonld",
        "https://w3id.org/vc/status-list/2021/v1": "status_list_context.jsonld",
        "https://www.w3.org/ns/did/v1": "did_documents_context.jsonld",
        "https://w3id.org/security/v1": "security-v1-context.jsonld",
        "https://w3id.org/security/v2": "security-v2-context.jsonld",
        "https://w3id.org/security/suites/ed25519-2020/v1": "ed25519-2020-context.jsonld",
        "https://w3id.org/security/bbs/v1": "bbs-v1-context.jsonld",
        "https://identity.foundation/presentation-exchange/submission/v1": (
            "dif-presentation-exchange-submission-v1.jsonld"
        ),
    }

    def __init__(
        self,
        document_downloader: Optional["JsonLdDocumentDownloader"] = None,
        document_parser: Optional["JsonLdDocumentParser"] = None,
    ):
        """Load static document on initialization."""
        self.documents_downloader = document_downloader or JsonLdDocumentDownloader()
        self.document_parser = document_parser or JsonLdDocumentParser()

        self.cache = {
            url: self.document_parser.parse(_load_jsonld_file(url, filename), None)
            for url, filename in StaticCacheJsonLdDownloader.CONTEXT_FILE_MAPPING.items()
        }

    def load(self, url: str, options: Optional[Dict] = None):
        """Load a jsonld document from URL.

        Prioritize local static cache before attempting to download from the URL.
        """
        cached = self.cache.get(url)

        if cached is not None:
            logger.info("Local cache hit for context: %s", url)
            return cached

        logger.debug("Context %s not in static cache, resolving from URL.", url)
        return self._live_load(url, options)

    def _live_load(self, url: str, options: Optional[Dict] = None):
        doc, link_header = self.documents_downloader.download(url, options)
        return self.document_parser.parse(doc, link_header)


class JsonLdDocumentDownloader:
    """JsonLd documents downloader."""

    def download(self, url: str, options: Optional[Dict], **kwargs):
        """Retrieves JSON-LD at the given URL.

        This was lifted from pyld.documentloader.requests.

        :param url: the URL to retrieve.
        :param options:

        :return: the RemoteDocument.

        """
        options = options or {}

        try:
            # validate URL
            pieces = urllib_parse.urlparse(url)
            if (
                not all([pieces.scheme, pieces.netloc])
                or pieces.scheme not in ["http", "https"]
                or set(pieces.netloc) > set(string.ascii_letters + string.digits + "-.:")
            ):
                raise JsonLdError(
                    'URL could not be dereferenced; only "http" and "https" '
                    "URLs are supported.",
                    "jsonld.InvalidUrl",
                    {"url": url},
                    code="loading document failed",
                )
            if options.get("secure") and pieces.scheme != "https":
                raise JsonLdError(
                    "URL could not be dereferenced; secure mode enabled and "
                    'the URL\'s scheme is not "https".',
                    "jsonld.InvalidUrl",
                    {"url": url},
                    code="loading document failed",
                )
            headers = options.get("headers")
            if headers is None:
                headers = {"Accept": "application/ld+json, application/json"}
            headers["User-Agent"] = f"AriesCloudAgent/{__version__}"
            response = requests.get(url, headers=headers, **kwargs)

            content_type = response.headers.get("content-type")
            if not content_type:
                content_type = "application/octet-stream"
            doc = {
                "contentType": content_type,
                "contextUrl": None,
                "documentUrl": response.url,
                "document": response.json(),
            }

            return doc, response.headers.get("link")
        except Exception as cause:
            raise JsonLdError(
                "Could not retrieve a JSON-LD document from the URL.",
                "jsonld.LoadDocumentError",
                code="loading document failed",
                cause=cause,
            )


class JsonLdDocumentParser:
    """JsonLd documents parser."""

    def parse(self, doc: Dict, link_header: Optional[str]):
        """Parse a jsonld document after retrieval.

        This was lifted from pyld.documentloader.requests.
        """
        try:
            if link_header:
                linked_context = parse_link_header(link_header).get(LINK_HEADER_REL)
                # only 1 related link header permitted
                if linked_context and doc["content_type"] != "application/ld+json":
                    if isinstance(linked_context, list):
                        raise JsonLdError(
                            "URL could not be dereferenced, "
                            "it has more than one "
                            "associated HTTP Link Header.",
                            "jsonld.LoadDocumentError",
                            {"url": doc["url"]},
                            code="multiple context link headers",
                        )
                    doc["contextUrl"] = linked_context["target"]
                linked_alternate = parse_link_header(link_header).get("alternate")
                # if not JSON-LD, alternate may point there
                if (
                    linked_alternate
                    and linked_alternate.get("type") == "application/ld+json"
                    and not re.match(r"^application\/(\w*\+)?json$", doc["content_type"])
                ):
                    doc["contentType"] = "application/ld+json"
                    doc["documentUrl"] = jsonld.prepend_base(
                        doc["url"], linked_alternate["target"]
                    )
            return doc
        except JsonLdError as e:
            raise e
        except Exception as cause:
            raise JsonLdError(
                "Could not retrieve a JSON-LD document from the URL.",
                "jsonld.LoadDocumentError",
                code="loading document failed",
                cause=cause,
            )



"""Base outbound transport."""

import asyncio
from abc import ABC, abstractmethod
from typing import Optional, Union

from ...connections.models.connection_target import ConnectionTarget
from ...core.profile import Profile
from ...utils.stats import Collector
from ..error import TransportError
from ..wire_format import BaseWireFormat
from .message import OutboundMessage


class QueuedOutboundMessage:
    """Class representing an outbound message pending delivery."""

    STATE_NEW = "new"
    STATE_PENDING = "pending"
    STATE_ENCODE = "encode"
    STATE_DELIVER = "deliver"
    STATE_RETRY = "retry"
    STATE_DONE = "done"

    def __init__(
        self,
        profile: Profile,
        message: OutboundMessage,
        target: ConnectionTarget,
        transport_id: str,
    ):
        """Initialize the queued outbound message."""
        self.profile = profile
        self.endpoint = target and target.endpoint
        self.error: Optional[Exception] = None
        self.message = message
        self.payload: Union[str, bytes] = None
        self.retries = None
        self.retry_at: Optional[float] = None
        self.state = self.STATE_NEW
        self.target = target
        self.task: asyncio.Task = None
        self.transport_id: str = transport_id
        self.metadata: Optional[dict] = None
        self.api_key: Optional[str] = None


class BaseOutboundTransport(ABC):
    """Base outbound transport class."""

    def __init__(
        self,
        wire_format: Optional[BaseWireFormat] = None,
        root_profile: Optional[Profile] = None,
    ) -> None:
        """Initialize a `BaseOutboundTransport` instance."""
        self._collector = None
        self._wire_format = wire_format
        self.root_profile = root_profile

    @property
    def collector(self) -> Collector:
        """Accessor for the stats collector instance."""
        return self._collector

    @collector.setter
    def collector(self, coll: Collector):
        """Assign a new stats collector instance."""
        self._collector = coll

    @property
    def wire_format(self) -> BaseWireFormat:
        """Accessor for a custom wire format for the transport."""
        return self._wire_format

    @wire_format.setter
    def wire_format(self, format: BaseWireFormat):
        """Setter for a custom wire format for the transport."""
        self._wire_format = format

    async def __aenter__(self):
        """Async context manager enter."""
        await self.start()

    async def __aexit__(self, err_type, err_value, err_t):
        """Async context manager exit."""
        if err_type and err_type != asyncio.CancelledError:
            self.logger.exception("Exception in outbound transport")
        await self.stop()

    @abstractmethod
    async def start(self):
        """Start the transport."""

    @abstractmethod
    async def stop(self):
        """Shut down the transport."""

    @abstractmethod
    async def handle_message(
        self,
        profile: Profile,
        outbound_message: QueuedOutboundMessage,
        endpoint: str,
        metadata: Optional[dict] = None,
    ):
        """Handle message.

        Args:
            profile: the profile that produced the message
            outbound_message: the outbound message to handle
            endpoint: URI endpoint for delivery
            metadata: Additional metadata associated with the payload
        """


class OutboundTransportError(TransportError):
    """Generic outbound transport error."""


class OutboundTransportRegistrationError(OutboundTransportError):
    """Outbound transport registration error."""


class OutboundDeliveryError(OutboundTransportError):
    """Base exception when a message cannot be delivered via an outbound transport."""



"""Protocol controller for issue credential v1_0."""

from typing import Sequence

PARTICIPATE_VC_INTERACTION = "aries.vc"
ISSUE_VC = "aries.vc.issue"


class Controller:
    """Issue credential v1_0 protocol controller."""

    def __init__(self, protocol: str):
        """Initialize the controller."""

    def determine_goal_codes(self) -> Sequence[str]:
        """Return defined goal_codes."""
        return [PARTICIPATE_VC_INTERACTION, ISSUE_VC]



"""Multikey class."""

import logging
from ...core.profile import ProfileSession
from ...resolver.did_resolver import DIDResolver
from ...utils.multiformats import multibase
from ...wallet.error import WalletError, WalletNotFoundError
from ..base import BaseWallet
from ..key_type import BLS12381G2, ED25519, P256, KeyType
from ..util import b58_to_bytes, bytes_to_b58
from pydid import VerificationMethod

LOGGER = logging.getLogger(__name__)

DEFAULT_ALG = "ed25519"
ALG_MAPPINGS = {
    "ed25519": {
        "key_type": ED25519,
        "multikey_prefix": "z6Mk",
        "prefix_hex": "ed01",
        "prefix_length": 2,
    },
    "x25519": {
        "key_type": ED25519,
        "multikey_prefix": "z6LS",
        "prefix_hex": "ec01",
        "prefix_length": 2,
    },
    "p256": {
        "key_type": P256,
        "multikey_prefix": "zDn",
        "prefix_hex": "8024",
        "prefix_length": 2,
    },
    "bls12381g2": {
        "key_type": BLS12381G2,
        "multikey_prefix": "zUC7",
        "prefix_hex": "eb01",
        "prefix_length": 2,
    },
}


def multikey_to_verkey(multikey: str):
    """Transform multikey to verkey."""

    alg = key_type_from_multikey(multikey).key_type
    prefix_length = ALG_MAPPINGS[alg]["prefix_length"]
    public_bytes = bytes(bytearray(multibase.decode(multikey))[prefix_length:])

    return bytes_to_b58(public_bytes)


def verkey_to_multikey(verkey: str, alg: str):
    """Transform verkey to multikey."""

    prefix_hex = ALG_MAPPINGS[alg]["prefix_hex"]
    prefixed_key_hex = f"{prefix_hex}{b58_to_bytes(verkey).hex()}"

    return multibase.encode(bytes.fromhex(prefixed_key_hex), "base58btc")


def key_type_from_multikey(multikey: str) -> KeyType:
    """Derive key_type class from multikey prefix."""
    for mapping in ALG_MAPPINGS:
        if multikey.startswith(ALG_MAPPINGS[mapping]["multikey_prefix"]):
            return ALG_MAPPINGS[mapping]["key_type"]

    raise MultikeyManagerError(f"Unsupported key algorithm for multikey {multikey}.")


def multikey_from_verification_method(verification_method: VerificationMethod) -> str:
    """Derive a multikey from a VerificationMethod."""
    if verification_method.type == "Multikey":
        multikey = verification_method.public_key_multibase

    elif verification_method.type == "Ed25519VerificationKey2018":
        multikey = verkey_to_multikey(
            verification_method.public_key_base58, alg="ed25519"
        )

    elif verification_method.type == "Ed25519VerificationKey2020":
        multikey = verification_method.public_key_multibase

    elif verification_method.type == "Bls12381G2Key2020":
        multikey = verkey_to_multikey(
            verification_method.public_key_base58, alg="bls12381g2"
        )
    # TODO address JsonWebKey based verification methods

    else:
        raise MultikeyManagerError("Unknown verification method type.")

    return multikey


class MultikeyManagerError(Exception):
    """Generic MultikeyManager Error."""


class MultikeyManager:
    """Class for managing wallet keys."""

    def __init__(self, session: ProfileSession):
        """Initialize the MultikeyManager."""

        self.session: ProfileSession = session
        self.wallet: BaseWallet = session.inject(BaseWallet)

    async def resolve_and_bind_kid(self, kid: str):
        """Fetch key if exists, otherwise resolve and bind it.

        This function is idempotent.
        """
        if await self.kid_exists(kid):
            LOGGER.info(f"kid {kid} already bound in storage, will not resolve.")
            return await self.from_kid(kid)
        else:
            multikey = await self.resolve_multikey_from_verification_method_id(kid)
            LOGGER.info(
                f"kid {kid} binding not found in storage, \
                binding to resolved multikey {multikey}."
            )
            return await self.update(multikey, kid)

    async def resolve_multikey_from_verification_method_id(self, kid: str):
        """Derive a multikey from the verification method ID."""
        resolver = self.session.inject(DIDResolver)
        verification_method = await resolver.dereference(
            profile=self.session.profile, did_url=kid
        )

        return multikey_from_verification_method(verification_method)

    def key_type_from_multikey(self, multikey: str) -> KeyType:
        """Derive key_type class from multikey prefix."""
        for mapping in ALG_MAPPINGS:
            if multikey.startswith(ALG_MAPPINGS[mapping]["multikey_prefix"]):
                return ALG_MAPPINGS[mapping]["key_type"]

        raise MultikeyManagerError(f"Unsupported key algorithm for multikey {multikey}.")

    async def kid_exists(self, kid: str):
        """Check if kid exists."""

        try:
            key = await self.wallet.get_key_by_kid(kid=kid)

            if key:
                return True
            return False

        except (WalletNotFoundError, AttributeError):
            return False

    async def multikey_exists(self, multikey: str):
        """Check if a multikey exists in the wallet."""

        try:
            key_info = await self.wallet.get_signing_key(
                verkey=multikey_to_verkey(multikey)
            )

            if key_info:
                return True
            return False

        except (WalletNotFoundError, AttributeError):
            return False

    async def from_kid(self, kid: str):
        """Fetch a single key."""

        try:
            key_info = await self.wallet.get_key_by_kid(kid=kid)

            return {
                "kid": key_info.kid,
                "multikey": verkey_to_multikey(
                    key_info.verkey, alg=key_info.key_type.key_type
                ),
            }
        except WalletError as err:
            LOGGER.error(err)
            return None

    async def from_multikey(self, multikey: str):
        """Fetch a single key."""

        key_info = await self.wallet.get_signing_key(verkey=multikey_to_verkey(multikey))

        return {
            "kid": key_info.kid,
            "multikey": verkey_to_multikey(
                key_info.verkey, alg=key_info.key_type.key_type
            ),
        }

    async def create(self, seed: str = None, kid: str = None, alg: str = DEFAULT_ALG):
        """Create a new key pair."""

        if alg not in ALG_MAPPINGS:
            raise MultikeyManagerError(
                f"Unknown key algorithm, use one of {list(ALG_MAPPINGS.keys())}."
            )

        if kid and await self.kid_exists(kid=kid):
            raise MultikeyManagerError(f"kid '{kid}' already exists in wallet.")

        key_type = ALG_MAPPINGS[alg]["key_type"]
        key_info = await self.wallet.create_key(key_type=key_type, seed=seed, kid=kid)

        return {
            "kid": key_info.kid,
            "multikey": verkey_to_multikey(key_info.verkey, alg=alg),
        }

    async def update(self, multikey: str, kid: str, unbind=False):
        """Bind or unbind a kid with a key pair."""
        (
            await self.unbind_key_id(multikey, kid)
            if unbind
            else await self.bind_key_id(multikey, kid)
        )

        return {"kid": kid, "multikey": multikey}

    async def bind_key_id(self, multikey: str, kid: str):
        """Bind a new key id to a key pair."""
        try:
            return await self.wallet.assign_kid_to_key(multikey_to_verkey(multikey), kid)
        except WalletError as err:
            LOGGER.error(err)
            raise MultikeyManagerError(err)

    async def unbind_key_id(self, multikey: str, kid: str):
        """Unbind a key id from a key pair."""
        try:
            return await self.wallet.unassign_kid_from_key(
                multikey_to_verkey(multikey), kid
            )
        except WalletError as err:
            LOGGER.error(err)
            raise MultikeyManagerError(err)



from unittest import TestCase, mock

import pytest

from .....messaging.models.base import BaseModelError
from ....didcomm_prefix import DIDCommPrefix
from ..message import ProblemReport, ProblemReportSchema
from ..message_types import PROBLEM_REPORT


class TestProblemReport(TestCase):
    """Problem report tests."""

    def test_init_type(self):
        """Test initializer."""

        prob = ProblemReport(
            description={
                "en": "oh no",
                "code": "abandoned",
            }
        )
        assert prob._type == DIDCommPrefix.qualify_current(PROBLEM_REPORT)

    def test_deserialize(self):
        """Test deserialization."""

        obj = ProblemReport(
            description={
                "en": "oh no",
                "code": "abandoned",
            }
        )

        with mock.patch.object(
            ProblemReportSchema, "load", mock.MagicMock()
        ) as mock_load:
            prob = ProblemReport.deserialize(obj)

        mock_load.assert_called_once_with(obj)
        assert prob is mock_load.return_value

    def test_serialize(self):
        """Test serialization."""

        obj = ProblemReport(
            description={
                "en": "oh no",
                "code": "abandoned",
            }
        )

        with mock.patch.object(
            ProblemReportSchema, "dump", mock.MagicMock()
        ) as mock_dump:
            ser = obj.serialize()

        mock_dump.assert_called_once_with(obj)
        assert ser is mock_dump.return_value

    def test_make_model(self):
        """Test making model."""

        prob = ProblemReport(
            description={
                "en": "oh no",
                "code": "abandoned",
            }
        )
        data = prob.serialize()
        model_instance = ProblemReport.deserialize(data)
        assert isinstance(model_instance, ProblemReport)

        prob = ProblemReport()
        data = prob.serialize()
        with pytest.raises(BaseModelError):
            ProblemReport.deserialize(data)



"""Ed25519Signature2018 suite."""

from datetime import datetime
from typing import Optional, Union

from ..crypto import _KeyPair as KeyPair
from .jws_linked_data_signature import JwsLinkedDataSignature


class Ed25519Signature2018(JwsLinkedDataSignature):
    """Ed25519Signature2018 suite."""

    signature_type = "Ed25519Signature2018"

    def __init__(
        self,
        *,
        key_pair: KeyPair,
        proof: Optional[dict] = None,
        verification_method: Optional[str] = None,
        date: Union[datetime, str, None] = None,
    ):
        """Create new Ed25519Signature2018 instance.

        Args:
            key_pair (KeyPair): Key pair to use. Must provide EdDSA signatures
            proof (dict, optional): A JSON-LD document with options to use for the
                `proof` node (e.g. any other custom fields can be provided here
                using a context different from security-v2).
            verification_method (str, optional): A key id URL to the paired public key.
            date (datetime, optional): Signing date to use.
        """
        super().__init__(
            algorithm="EdDSA",
            required_key_type="Ed25519VerificationKey2018",
            key_pair=key_pair,
            verification_method=verification_method,
            proof=proof,
            date=date,
        )



"""Message and inner object type identifiers for Connections."""

from ...didcomm_prefix import DIDCommPrefix
from .messages.cred_format import V20CredFormat

SPEC_URI = (
    "https://github.com/decentralized-identity/aries-rfcs/tree/"
    "cd27fc64aa2805f756a118043d7c880354353047/features/0453-issue-credential-v2"
)

# Message types
CRED_20_PROPOSAL = "issue-credential/2.0/propose-credential"
CRED_20_OFFER = "issue-credential/2.0/offer-credential"
CRED_20_REQUEST = "issue-credential/2.0/request-credential"
CRED_20_ISSUE = "issue-credential/2.0/issue-credential"
CRED_20_ACK = "issue-credential/2.0/ack"
CRED_20_PROBLEM_REPORT = "issue-credential/2.0/problem-report"

PROTOCOL_PACKAGE = "acapy_agent.protocols.issue_credential.v2_0"

MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {
        CRED_20_PROPOSAL: (f"{PROTOCOL_PACKAGE}.messages.cred_proposal.V20CredProposal"),
        CRED_20_OFFER: f"{PROTOCOL_PACKAGE}.messages.cred_offer.V20CredOffer",
        CRED_20_REQUEST: f"{PROTOCOL_PACKAGE}.messages.cred_request.V20CredRequest",
        CRED_20_ISSUE: f"{PROTOCOL_PACKAGE}.messages.cred_issue.V20CredIssue",
        CRED_20_ACK: f"{PROTOCOL_PACKAGE}.messages.cred_ack.V20CredAck",
        CRED_20_PROBLEM_REPORT: (
            f"{PROTOCOL_PACKAGE}.messages.cred_problem_report.V20CredProblemReport"
        ),
    }
)

# Inner object types
CRED_20_PREVIEW = "issue-credential/2.0/credential-preview"

# Format specifications
ATTACHMENT_FORMAT = {
    CRED_20_PROPOSAL: {
        V20CredFormat.Format.ANONCREDS.api: "anoncreds/credential-filter@v1.0",
        V20CredFormat.Format.INDY.api: "hlindy/cred-filter@v2.0",
        V20CredFormat.Format.LD_PROOF.api: "aries/ld-proof-vc-detail@v1.0",
        V20CredFormat.Format.VC_DI.api: "didcomm/w3c-di-vc@v0.1",
    },
    CRED_20_OFFER: {
        V20CredFormat.Format.ANONCREDS.api: "anoncreds/credential-offer@v1.0",
        V20CredFormat.Format.INDY.api: "hlindy/cred-abstract@v2.0",
        V20CredFormat.Format.LD_PROOF.api: "aries/ld-proof-vc-detail@v1.0",
        V20CredFormat.Format.VC_DI.api: "didcomm/w3c-di-vc-offer@v0.1",
    },
    CRED_20_REQUEST: {
        V20CredFormat.Format.ANONCREDS.api: "anoncreds/credential-request@v1.0",
        V20CredFormat.Format.INDY.api: "hlindy/cred-req@v2.0",
        V20CredFormat.Format.LD_PROOF.api: "aries/ld-proof-vc-detail@v1.0",
        V20CredFormat.Format.VC_DI.api: "didcomm/w3c-di-vc-request@v0.1",
    },
    CRED_20_ISSUE: {
        V20CredFormat.Format.ANONCREDS.api: "anoncreds/credential@v1.0",
        V20CredFormat.Format.INDY.api: "hlindy/cred@v2.0",
        V20CredFormat.Format.LD_PROOF.api: "aries/ld-proof-vc@v1.0",
        V20CredFormat.Format.VC_DI.api: "didcomm/w3c-di-vc@v0.1",
    },
}

CONTROLLERS = DIDCommPrefix.qualify_all(
    {"issue-credential/2.0": f"{PROTOCOL_PACKAGE}.controller.Controller"}
)



from unittest import IsolatedAsyncioTestCase, mock

from ......did.did_key import DIDKey
from ......wallet.key_type import ED25519
from .....out_of_band.v1_0.messages.invitation import InvitationMessage, Service
from ...message_types import PROTOCOL_PACKAGE
from ..invitation import Invitation as IntroInvitation


class TestInvitation(IsolatedAsyncioTestCase):
    def setUp(self):
        self.label = "Label"
        self.test_did = "55GkHamhTU1ZbTbV2ab9DE"
        self.did = "did:sov:QmWbsNYhMrjHiqZDTUTEJs"
        self.endpoint_url = "https://example.com/endpoint"
        self.endpoint_did = "did:sov:A2wBhNYhMrjHiqZDTUYH7u"
        self.key = "8HH5gYEeNc3z7PYXmd54d4x6qAfCNrqQqEB3nS7Zfu7K"
        self.did_key = DIDKey.from_public_key_b58(self.key, ED25519)
        self.test_message = "test message"

        self.service = Service(
            _id="asdf",
            _type="did-communication",
            recipient_keys=[self.did_key.did],
            service_endpoint=self.endpoint_url,
        )
        self.conn_invi_msg = InvitationMessage(
            label=self.label,
            services=[self.service],
            handshake_protocols=["didexchange/1.1"],
        )
        self.intro_invitation = IntroInvitation(
            invitation=self.conn_invi_msg,
            message=self.test_message,
        )

    def test_init(self):
        """Test initialization."""
        assert self.intro_invitation.invitation == self.conn_invi_msg
        assert self.intro_invitation.message == self.test_message

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.invitation.InvitationSchema.load")
    def test_deserialize(self, mock_invitation_schema_load):
        """
        Test deserialization.
        """
        obj = {"obj": "obj"}

        intro_invi = IntroInvitation.deserialize(obj)
        mock_invitation_schema_load.assert_called_once_with(obj)

        assert intro_invi is mock_invitation_schema_load.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.invitation.InvitationSchema.dump")
    def test_serialize(self, mock_invitation_schema_dump):
        """
        Test serialization.
        """
        intro_invi_dict = self.intro_invitation.serialize()
        mock_invitation_schema_dump.assert_called_once_with(self.intro_invitation)

        assert intro_invi_dict is mock_invitation_schema_dump.return_value

    async def test_make_model(self):
        intro_invi = IntroInvitation(
            invitation=self.conn_invi_msg,
            message=self.test_message,
        )

        data = intro_invi.serialize()
        model_instance = IntroInvitation.deserialize(data)
        assert type(model_instance) is type(intro_invi)
